#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from pyasn1.compat.octets import str2octs


# read lines from text file ignoring #comments and blank lines
def getRecord(fileObj, lineNo=None, offset=0):
    line = fileObj.readline()
    if lineNo is not None and line:
        lineNo += 1
    while line:
        tline = line.strip()
        # skip comment or blank line
        if not tline or tline.startswith(str2octs('#')):
            offset += len(line)
            line = fileObj.readline()
            if lineNo is not None and line:
                lineNo += 1
        else:
            break
    return line, lineNo, offset


def findEol(fileObj, offset, blockSize=256, eol=str2octs('\n')):
    while True:
        if offset < blockSize:
            offset, blockSize = 0, offset
        else:
            offset -= blockSize
        fileObj.seek(offset)
        chunk = fileObj.read(blockSize)
        try:
            return offset + chunk.rindex(eol) + 1
        except ValueError:
            if offset == 0:
                return offset
            continue


# In-place, by-OID binary search
def searchRecordByOid(oid, fileObj, textParser):
    lo = mid = 0;
    prev_mid = -1
    fileObj.seek(0, 2)
    hi = sz = fileObj.tell()
    while lo < hi:
        mid = (lo + hi) // 2
        fileObj.seek(mid)
        mid = findEol(fileObj, mid)
        fileObj.seek(mid)
        if mid == prev_mid:  # loop condition due to stepping back pivot
            break
        if mid >= sz:
            return sz
        line, _, skippedOffset = getRecord(fileObj)
        if not line:
            return hi
        midval, _ = textParser.evaluate(line, oidOnly=True)
        if midval < oid:
            lo = mid + skippedOffset + len(line)
        elif midval > oid:
            hi = mid
        else:
            return mid
        prev_mid = mid
    if lo == mid:
        return lo
    else:
        return hi



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim import log
from snmpsim.mltsplit import split
from pysnmp.smi import error


errorTypes = {
    'toobig': error.TooBigError,
    'nosuchname': error.NoSuchNameError,
    'badvalue': error.BadValueError,
    'readonly': error.ReadOnlyError,
    'generror': error.GenError,
    'noaccess': error.NoAccessError,
    'wrongtype': error.WrongTypeError,
    'wrongvalue': error.WrongValueError,
    'nocreation': error.NoCreationError,
    'inconsistentvalue': error.InconsistentValueError,
    'resourceunavailable': error.ResourceUnavailableError,
    'commitfailed': error.CommitFailedError,
    'undofailed': error.UndoFailedError,
    'authorizationerror': error.AuthorizationError,
    'notwritable': error.NotWritableError,
    'inconsistentname': error.InconsistentNameError,
    # These only makes sense for SNMP v1
    'nosuchobject': error.NoSuchObjectError,
    'nosuchinstance': error.NoSuchInstanceError,
    'endofmib': error.EndOfMibViewError
}


def init(**context):
    pass


def variate(oid, tag, value, **context):
    if not context['nextFlag'] and not context['exactMatch']:
        return context['origOid'], tag, context['errorStatus']

    if 'settings' not in recordContext:
        recordContext['settings'] = dict([split(x, '=') for x in split(value, ',')])

        if 'hexvalue' in recordContext['settings']:
            recordContext['settings']['value'] = [int(recordContext['settings']['hexvalue'][x:x + 2], 16) for x in
                                                  range(0, len(recordContext['settings']['hexvalue']), 2)]

        if 'status' in recordContext['settings']:
            recordContext['settings']['status'] = recordContext['settings']['status'].lower()

        if 'op' not in recordContext['settings']:
            recordContext['settings']['op'] = 'any'

        if 'vlist' in recordContext['settings']:
            vlist = {}
            recordContext['settings']['vlist'] = split(recordContext['settings']['vlist'], ':')
            while recordContext['settings']['vlist']:
                o, v, e = recordContext['settings']['vlist'][:3]
                recordContext['settings']['vlist'] = recordContext['settings']['vlist'][3:]
                typeTag, _ = SnmprecRecord.unpackTag(tag)
                v = SnmprecGrammar.tagMap[typeTag](v)
                if o not in vlist:
                    vlist[o] = {}
                if o == 'eq':
                    vlist[o][v] = e
                elif o in ('lt', 'gt'):
                    vlist[o] = v, e
                else:
                    log.msg('error: bad vlist syntax: %s' % recordContext['settings']['vlist'])
            recordContext['settings']['vlist'] = vlist

    e = None

    if context['setFlag']:
        if 'vlist' in recordContext['settings']:
            if ('eq' in recordContext['settings']['vlist'] and
                    context['origValue'] in recordContext['settings']['vlist']['eq']):
                e = recordContext['settings']['vlist']['eq'][context['origValue']]
            elif ('lt' in recordContext['settings']['vlist'] and
                    context['origValue'] < recordContext['settings']['vlist']['lt'][0]):
                e = recordContext['settings']['vlist']['lt'][1]
            elif ('gt' in recordContext['settings']['vlist'] and
                    context['origValue'] > recordContext['settings']['vlist']['gt'][0]):
                e = recordContext['settings']['vlist']['gt'][1]
        elif recordContext['settings']['op'] in ('set', 'any'):
            if 'status' in recordContext['settings']:
                e = recordContext['settings']['status']
    else:
        if recordContext['settings']['op'] in ('get', 'any'):
            if 'status' in recordContext['settings']:
                e = recordContext['settings']['status']

    if e and e in errorTypes:
        log.msg('error: reporting %s for %s' % (e, oid))
        raise errorTypes[e](
            name=oid, idx=max(0, context['varsTotal'] - context['varsRemaining'] - 1)
        )

    if context['setFlag']:
        recordContext['settings']['value'] = context['origValue']

    return oid, tag, recordContext['settings'].get('value', context['errorStatus'])


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from snmpsim.grammar import abstract
from snmpsim.error import SnmpsimError


class AbstractRecord:
    grammar = abstract.AbstractGrammar()
    ext = ''

    def evaluateOid(self, oid):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)

    def evaluateValue(self, oid, tag, value, **context):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)

    def evaluate(self, line, **context):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)

    def formatOid(self, oid):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)

    def formatValue(self, oid, value, **context):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)

    def format(self, oid, value, **context):
        raise SnmpsimError('Method not implemented at %s' % self.__class__.__name__)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# Managed value variation module
# Send SNMP Notification
#
from pysnmp.hlapi.asyncore import *
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim.mltsplit import split
from snmpsim import error, log


def init(**context):
    pass


typeMap = {
    's': OctetString,
    'h': lambda x: OctetString(hexValue=x),
    'i': Integer32,
    'o': ObjectIdentifier,
    'a': IpAddress,
    'u': Unsigned32,
    'g': Gauge32,
    't': TimeTicks,
    'b': Bits,
    'I': Counter64
}


def _cbFun(sendRequestHandle,
           errorIndication,
           errorStatus, errorIndex,
           varBinds,
           cbCtx):
    oid, value = cbCtx
    if errorIndication or errorStatus:
        log.msg('notification: for %s=%r failed with errorIndication %s, errorStatus %s' % (oid, value, errorIndication, errorStatus))


def variate(oid, tag, value, **context):
    if 'snmpEngine' in context and context['snmpEngine']:
        snmpEngine = context['snmpEngine']
        if snmpEngine not in moduleContext:
            moduleContext[snmpEngine] = {}
        if context['transportDomain'] not in moduleContext[snmpEngine]:
            # register this SNMP Engine to handle our transports'
            # receiver IDs (which we build by outbound and simulator
            # transportDomains concatenation)
            snmpEngine.registerTransportDispatcher(
                snmpEngine.transportDispatcher,
                UdpTransportTarget.transportDomain + context['transportDomain']
            )
            snmpEngine.registerTransportDispatcher(
                snmpEngine.transportDispatcher,
                Udp6TransportTarget.transportDomain + context['transportDomain']
            )
            moduleContext[snmpEngine][context['transportDomain']] = 1
    else:
        raise error.SnmpsimError(
            'Variation module is not given snmpEngine. '
            'Make sure you are not running in --v2c-arch mode')

    if not context['nextFlag'] and not context['exactMatch']:
        return context['origOid'], tag, context['errorStatus']

    if 'settings' not in recordContext:
        recordContext['settings'] = dict([split(x, '=') for x in split(value, ',')])
        for k, v in (('op', 'set'),
                     ('community', 'public'),
                     ('authkey', None),
                     ('authproto', 'md5'),
                     ('privkey', None),
                     ('privproto', 'des'),
                     ('proto', 'udp'),
                     ('port', '162'),
                     ('ntftype', 'trap'),
                     ('trapoid', '1.3.6.1.6.3.1.1.5.1')):
            recordContext['settings'].setdefault(k, v)

        if 'hexvalue' in recordContext['settings']:
            recordContext['settings']['value'] = [int(recordContext['settings']['hexvalue'][x:x + 2], 16) for x in
                                                  range(0, len(recordContext['settings']['hexvalue']), 2)]

        if 'vlist' in recordContext['settings']:
            vlist = {}
            recordContext['settings']['vlist'] = split(recordContext['settings']['vlist'], ':')
            while recordContext['settings']['vlist']:
                o, v = recordContext['settings']['vlist'][:2]
                recordContext['settings']['vlist'] = recordContext['settings']['vlist'][2:]
                typeTag, _ = SnmprecRecord.unpackTag(tag)
                v = SnmprecGrammar.tagMap[typeTag](v)
                if o not in vlist:
                    vlist[o] = set()
                if o == 'eq':
                    vlist[o].add(v)
                elif o in ('lt', 'gt'):
                    vlist[o] = v
                else:
                    log.msg('notification: bad vlist syntax: %s' % recordContext['settings']['vlist'])
            recordContext['settings']['vlist'] = vlist

    args = recordContext['settings']

    if context['setFlag'] and 'vlist' in args:
        if ('eq' in args['vlist'] and
                context['origValue'] in args['vlist']['eq']):
            pass
        elif ('lt' in args['vlist'] and
                context['origValue'] < args['vlist']['lt']):
            pass
        elif ('gt' in args['vlist'] and
                context['origValue'] > args['vlist']['gt']):
            pass
        else:
            return oid, tag, context['origValue']

    if args['op'] not in ('get', 'set', 'any', '*'):
        log.msg('notification: unknown SNMP request type configured: %s' % args['op'])
        return context['origOid'], tag, context['errorStatus']

    if (args['op'] == 'get' and not context['setFlag'] or
                args['op'] == 'set' and context['setFlag'] or
                args['op'] in ('any', '*')):
        if args['version'] in ('1', '2c'):
            authData = CommunityData(args['community'], mpModel=args['version'] == '2c' and 1 or 0)
        elif args['version'] == '3':
            if args['authproto'] == 'md5':
                authProtocol = usmHMACMD5AuthProtocol
            elif args['authproto'] == 'sha':
                authProtocol = usmHMACSHAAuthProtocol
            elif args['authproto'] == 'none':
                authProtocol = usmNoAuthProtocol
            else:
                log.msg('notification: unknown auth proto %s' % args['authproto'])
                return context['origOid'], tag, context['errorStatus']
            if args['privproto'] == 'des':
                privProtocol = usmDESPrivProtocol
            elif args['privproto'] == 'aes':
                privProtocol = usmAesCfb128Protocol
            elif args['privproto'] == 'none':
                privProtocol = usmNoPrivProtocol
            else:
                log.msg('notification: unknown privacy proto %s' % args['privproto'])
                return context['origOid'], tag, context['errorStatus']
            authData = UsmUserData(args['user'], args['authkey'], args['privkey'], authProtocol=authProtocol,
                                   privProtocol=privProtocol)
        else:
            log.msg('notification: unknown SNMP version %s' % args['version'])
            return context['origOid'], tag, context['errorStatus']

        if 'host' not in args:
            log.msg('notification: target hostname not configured for OID %s' % (oid,))
            return context['origOid'], tag, context['errorStatus']

        if args['proto'] == 'udp':
            target = UdpTransportTarget((args['host'], int(args['port'])))
        elif args['proto'] == 'udp6':
            target = Udp6TransportTarget((args['host'], int(args['port'])))
        else:
            log.msg('notification: unknown transport %s' % args['proto'])
            return context['origOid'], tag, context['errorStatus']

        localAddress = None

        if 'bindaddr' in args:
            localAddress = args['bindaddr']
        else:
            if context['transportDomain'][:len(target.transportDomain)] == target.transportDomain:
                localAddress = snmpEngine.transportDispatcher.getTransport(context['transportDomain']).getLocalAddress()[0]
            else:
                log.msg('notification: incompatible network transport types used by CommandResponder vs NotificationOriginator')
                if 'bindaddr' in args:
                    localAddress = args['bindaddr']

        if localAddress:
            log.msg('notification: binding to local address %s' % localAddress)
            target.setLocalAddress((localAddress, 0))

        # this will make target objects different based on their bind address 
        target.transportDomain = target.transportDomain + context['transportDomain']

        varBinds = []

        if 'uptime' in args:
            varBinds.append(
                (ObjectIdentifier('1.3.6.1.2.1.1.3.0'),
                 TimeTicks(args['uptime']))
            )

        if args['version'] == '1':
            if 'agentaddress' in args:
                varBinds.append(
                    (ObjectIdentifier('1.3.6.1.6.3.18.1.3.0'),
                     IpAddress(args['agentaddress']))
                )
            if 'enterprise' in args:
                varBinds.append(
                    (ObjectIdentifier('1.3.6.1.6.3.1.1.4.3.0'),
                     ObjectIdentifier(args['enterprise']))
                )

        if 'varbinds' in args:
            vbs = split(args['varbinds'], ':')
            while vbs:
                varBinds.append(
                    (ObjectIdentifier(vbs[0]), typeMap[vbs[1]](vbs[2]))
                )
                vbs = vbs[3:]

        sendNotification(
            snmpEngine, authData, target, ContextData(), args['ntftype'],
            NotificationType(ObjectIdentity(args['trapoid'])).addVarBinds(*varBinds),
            cbFun=_cbFun, cbCtx=(oid, value)
        )

        log.msg('notification: sending Notification to %s with credentials %s' % (authData, target))

    if context['setFlag'] or 'value' not in args:
        return oid, tag, context['origValue']
    else:
        return oid, tag, args['value']


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# Managed value variation module
# Get/set managed value by invoking an external program
#
import sys
import subprocess
from pysnmp.proto import rfc1902
from snmpsim.mltsplit import split
from snmpsim import log


def init(**context):
    moduleContext['settings'] = {}
    if context['options']:
        moduleContext['settings'].update(
            dict([split(x, ':') for x in split(context['options'], ',')])
        )
    if 'shell' not in moduleContext['settings']:
        moduleContext['settings']['shell'] = sys.platform[:3] == 'win'
    else:
        moduleContext['settings']['shell'] = int(moduleContext['settings']['shell'])


def variate(oid, tag, value, **context):
    # in --v2c-arch some of the items are not defined
    transportDomain = transportAddress = securityModel = securityName = \
        securityLevel = contextName = '<undefined>'
    if 'transportDomain' in context:
        transportDomain = rfc1902.ObjectName(context['transportDomain']).prettyPrint()
    if 'transportAddress' in context:
        transportAddress = ':'.join([str(x) for x in context['transportAddress']])
    if 'securityModel' in context:
        securityModel = str(context['securityModel'])
    if 'securityName' in context:
        securityName = str(context['securityName'])
    if 'securityLevel' in context:
        securityLevel = str(context['securityLevel'])
    if 'contextName' in context:
        contextName = str(context['contextName'])

    args = [(x
            .replace('@TRANSPORTDOMAIN@', transportDomain)
            .replace('@TRANSPORTADDRESS@', transportAddress)
            .replace('@SECURITYMODEL@', securityModel)
            .replace('@SECURITYNAME@', securityName)
            .replace('@SECURITYLEVEL@', securityLevel)
            .replace('@CONTEXTNAME@', contextName)
            .replace('@DATAFILE@', context['dataFile'])
            .replace('@OID@', str(oid))
            .replace('@TAG@', tag)
            .replace('@ORIGOID@', str(context['origOid']))
            .replace('@ORIGTAG@', str(sum([x for x in context['origValue'].tagSet[0]])))
            .replace('@ORIGVALUE@', str(context['origValue']))
            .replace('@SETFLAG@', str(int(context['setFlag'])))
            .replace('@NEXTFLAG@', str(int(context['nextFlag'])))
            .replace('@SUBTREEFLAG@', str(int(context['subtreeFlag']))))
            for x in split(value, ' ')]

    log.msg('subprocess: executing external process "%s"' % ' '.join(args))

    call = (hasattr(subprocess, 'check_output') and subprocess.check_output or
            hasattr(subprocess, 'check_call') and subprocess.check_call or
            subprocess.call)
    if not hasattr(subprocess, 'check_output'):
        log.msg('subprocess: old Python, expect no output!')

    try:
        return oid, tag, call(
            args, shell=moduleContext['settings']['shell']
        )
    except (hasattr(subprocess, 'CalledProcessError') and
            subprocess.CalledProcessError or Exception):
        log.msg('subprocess: external program execution failed')
        return context['origOid'], tag, context['errorStatus']


def shutdown(**context):
    pass



#!/usr/bin/env python
#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Simulator MIB to data file converter
#
import getopt
import sys
import random
from pyasn1.type import univ
from pyasn1.error import PyAsn1Error
from pysnmp.smi import compiler, builder, view, error
from pysnmp.smi.rfc1902 import ObjectIdentity
from pysnmp.proto import rfc1902
from pysnmp import debug
from snmpsim.record import snmprec
from snmpsim.error import SnmpsimError

# Defaults
verboseFlag = True
startOID = stopOID = None
mibSources = []
defaultMibSources = ['http://mibs.snmplabs.com/asn1/@mib@']
outputFile = sys.stdout
if hasattr(outputFile, 'buffer'):
    outputFile = outputFile.buffer
stringPool = 'Jaded zombies acted quaintly but kept driving their oxen forward'.split()
counterRange = (0, 0xffffffff)
counter64Range = (0, 0xffffffffffffffff)
gaugeRange = (0, 0xffffffff)
timeticksRange = (0, 0xffffffff)
unsignedRange = (0, 65535)
int32Range = (0, 32)  # these values are more likely to fit constraints
automaticValues = 5000
tableSize = 10
modNames = []
mibDirs = []

helpMessage = """Usage: %s [--help]
    [--version]
    [--debug=<%s>]
    [--quiet]
    [--mib-source=<url>]
    [--pysnmp-mib-dir=</path/to/pysnmp/mibs]
    [--mib-module=<MIB-NAME>]
    [--start-object=<MIB-NAME::[symbol-name]|OID>]
    [--stop-object=<MIB-NAME::[symbol-name]|OID>]
    [--manual-values]
    [--automatic-values=<max-probes>]
    [--table-size=<number>]
    [--output-file=<filename>]
    [--string-pool=<words>]
    [--string-pool-file=</path/to/text/file>]
    [--counter-range=<min,max>]
    [--counter64-range=<min,max>]
    [--gauge-range=<min,max>]
    [--timeticks-range=<min,max>]
    [--unsigned-range=<min,max>]
    [--integer32-range=<min,max>]""" % (
    sys.argv[0],
    '|'.join([x for x in getattr(debug, 'FLAG_MAP', getattr(debug, 'flagMap', ()))
              if x not in ('app', 'msgproc', 'proxy', 'io', 'secmod', 'dsp', 'acl')])
)

try:
    opts, params = getopt.getopt(sys.argv[1:], 'hv', [
        'help', 'version', 'debug=', 'quiet',
        'pysnmp-mib-dir=', 'mib-module=', 'start-oid=', 'stop-oid=',
        'mib-source=', 'start-object=', 'stop-object=',
        'manual-values', 'automatic-values=', 'table-size=',
        'output-file=', 'string-pool=', 'string-pool-file=', 'integer32-range=',
        'counter-range=', 'counter64-range=', 'gauge-range=',
        'unsigned-range=', 'timeticks-range='
    ])
except Exception:
    sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
    sys.exit(-1)

if params:
    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
    sys.exit(-1)

for opt in opts:
    if opt[0] == '-h' or opt[0] == '--help':
        sys.stderr.write("""\
Synopsis:
  Converts MIB modules into SNMP Simulator data files.
  Chooses random values or can ask for them interactively.
  Able to fill SNMP conceptual tables with consistent indices.
Documentation:
  http://snmplabs.com/snmpsim/simulation-based-on-mibs.html
%s
""" % helpMessage)
        sys.exit(-1)
    if opt[0] == '-v' or opt[0] == '--version':
        import snmpsim
        import pysmi
        import pysnmp
        import pyasn1

        sys.stderr.write("""\
SNMP Simulator version %s, written by Ilya Etingof <etingof@gmail.com>
Using foundation libraries: pysmi %s, pysnmp %s, pyasn1 %s.
Python interpreter: %s
Software documentation and support at http://snmplabs.com/snmpsim
%s
""" % (snmpsim.__version__,
       hasattr(pysmi, '__version__') and pysmi.__version__ or 'unknown',
       hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown',
       hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown',
       sys.version, helpMessage))
        sys.exit(-1)
    if opt[0] == '--debug':
        debug.setLogger(debug.Debug(*opt[1].split(',')))
    if opt[0] == '--quiet':
        verboseFlag = False
    if opt[0] == '--pysnmp-mib-dir':
        mibDirs.append(opt[1])
    if opt[0] == '--mib-module':
        modNames.append(opt[1])
    # obsolete begin
    if opt[0] == '--start-oid':
        startOID = univ.ObjectIdentifier(opt[1])
    if opt[0] == '--stop-oid':
        stopOID = univ.ObjectIdentifier(opt[1])
    # obsolete end
    if opt[0] == '--mib-source':
        mibSources.append(opt[1])
    if opt[0] == '--start-object':
        startOID = ObjectIdentity(*opt[1].split('::', 1))
    if opt[0] == '--stop-object':
        stopOID = ObjectIdentity(*opt[1].split('::', 1), **dict(last=True))
    if opt[0] == '--manual-values':
        automaticValues = 0
    if opt[0] == '--automatic-values':
        try:
            automaticValues = int(opt[1])
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--table-size':
        try:
            tableSize = int(opt[1])
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--output-file':
        outputFile = open(opt[1], 'wb')
    if opt[0] == '--string-pool':
        stringPool = opt[1].split()
    if opt[0] == '--string-pool-file':
        try:
            f = open(opt[1])
            stringPool = f.read().split()
            f.close()
        except Exception:
            sys.stderr.write('ERROR: text file "%s" open failure %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--counter-range':
        try:
            counterRange = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write(
                'ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--counter64-range':
        try:
            counter64Range = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--gauge-range':
        try:
            gaugeRange = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--timeticks-range':
        try:
            timeticksRange = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--integer32-range':
        try:
            int32Range = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)
    if opt[0] == '--unsigned-range':
        try:
            unsignedRange = [int(x) for x in opt[1].split(',')]
        except ValueError:
            sys.stderr.write('ERROR: bad value %s: %s\r\n' % (opt[1], sys.exc_info()[1]))
            sys.exit(-1)

# Catch missing params
if not modNames:
    sys.stderr.write('ERROR: MIB modules not specified\r\n%s\r\n' % helpMessage)
    sys.exit(-1)


def getValue(syntax, hint='', automaticValues=automaticValues):
    # Optionally approve chosen value with the user
    makeGuess = automaticValues
    val = None
    while True:
        if makeGuess:
            # Pick a value
            if isinstance(syntax, rfc1902.IpAddress):
                val = '.'.join([str(random.randrange(1, 256)) for x in range(4)])
            elif isinstance(syntax, rfc1902.TimeTicks):
                val = random.randrange(timeticksRange[0], timeticksRange[1])
            elif isinstance(syntax, rfc1902.Gauge32):
                val = random.randrange(gaugeRange[0], gaugeRange[1])
            elif isinstance(syntax, rfc1902.Counter32):
                val = random.randrange(counterRange[0], counterRange[1])
            elif isinstance(syntax, rfc1902.Integer32):
                val = random.randrange(int32Range[0], int32Range[1])
            elif isinstance(syntax, rfc1902.Unsigned32):
                val = random.randrange(unsignedRange[0], unsignedRange[1])
            elif isinstance(syntax, rfc1902.Counter64):
                val = random.randrange(counter64Range[0], counter64Range[1])
            elif isinstance(syntax, univ.OctetString):
                maxWords = 10
                val = ' '.join([stringPool[random.randrange(0, len(stringPool))] for i in range(random.randrange(1, maxWords))])
            elif isinstance(syntax, univ.ObjectIdentifier):
                val = '.'.join(['1', '3', '6', '1', '3'] + [
                    '%d' % random.randrange(0, 255) for x in range(random.randrange(0, 10))])
            elif isinstance(syntax, rfc1902.Bits):
                val = [random.randrange(0, 256) for x in range(random.randrange(0, 9))]
            else:
                val = '?'

        try:
            # remove value enumeration

            if syntax.tagSet == rfc1902.Integer32.tagSet:
                return rfc1902.Integer32(syntax.clone(val))

            if syntax.tagSet == rfc1902.Unsigned32.tagSet:
                return rfc1902.Unsigned32(syntax.clone(val))

            if syntax.tagSet == rfc1902.Bits.tagSet:
                return rfc1902.OctetString(syntax.clone(val))

            return syntax.clone(val)

        except PyAsn1Error:
            if makeGuess == 1:
                sys.stderr.write(
                    '*** Inconsistent value: %s\r\n*** See constraints and suggest a better one for:\r\n' % (sys.exc_info()[1],)
                )
            if makeGuess:
                makeGuess -= 1
                continue

        sys.stderr.write('%s# Value [\'%s\'] ? ' % (hint, (val is None and '<none>' or val),))
        sys.stderr.flush()
        line = sys.stdin.readline().strip()
        if line:
            if line[:2] == '0x':
                if line[:4] == '0x0x':
                    line = line[2:]
                elif isinstance(syntax, univ.OctetString):
                    val = syntax.clone(hexValue=line[2:])
                else:
                    val = int(line[2:], 16)
            else:
                val = line


# Data file builder

dataFileHandler = snmprec.SnmprecRecord()

mibBuilder = builder.MibBuilder()

mibBuilder.setMibSources(
    *mibBuilder.getMibSources() + tuple(
        [builder.ZipMibSource(m).init() for m in mibDirs]
    )
)

# Load MIB tree foundation classes
(MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn) = mibBuilder.importSymbols(
    'SNMPv2-SMI',
    'MibScalar',
    'MibTable',
    'MibTableRow',
    'MibTableColumn'
)

mibViewController = view.MibViewController(mibBuilder)

compiler.addMibCompiler(
    mibBuilder, sources=mibSources or defaultMibSources
)
if isinstance(startOID, ObjectIdentity):
    startOID.resolveWithMib(mibViewController)
if isinstance(stopOID, ObjectIdentity):
    stopOID.resolveWithMib(mibViewController)

output = []

# MIBs walk
for modName in modNames:
    if verboseFlag:
        sys.stderr.write('# MIB module: %s, from %s till %s\r\n' % (modName, startOID or 'the beginning', stopOID or 'the end'))
    oid = ObjectIdentity(modName).resolveWithMib(mibViewController)
    hint = rowHint = ''
    rowOID = None
    suffix = ()
    thisTableSize = 0
    while True:
        try:
            oid, label, _ = mibViewController.getNextNodeName(oid)
        except error.NoSuchObjectError:
            break

        if rowOID and not rowOID.isPrefixOf(oid):
            thisTableSize += 1
            if automaticValues:
                if thisTableSize < tableSize:
                    oid = tuple(rowOID)
                    if verboseFlag:
                        sys.stderr.write('# Synthesizing row #%d of table %s\r\n' % (thisTableSize, rowOID))
                else:
                    if verboseFlag:
                        sys.stderr.write('# Finished table %s (%d rows)\r\n' % (rowOID, thisTableSize))
                    rowOID = None
            else:
                while True:
                    sys.stderr.write('# Synthesize row #%d for table %s (y/n)? ' % (thisTableSize, rowOID))
                    sys.stderr.flush()
                    line = sys.stdin.readline().strip()
                    if line:
                        if line[0] in ('y', 'Y'):
                            oid = tuple(rowOID)
                            break
                        elif line[0] in ('n', 'N'):
                            if verboseFlag:
                                sys.stderr.write('# Finished table %s (%d rows)\r\n' % (rowOID, thisTableSize))
                            rowOID = None
                            break

        if startOID and oid < startOID:
            continue  # skip on premature OID
        if stopOID and oid > stopOID:
            break  # stop on out of range condition

        mibName, symName, _ = mibViewController.getNodeLocation(oid)
        node, = mibBuilder.importSymbols(mibName, symName)

        if isinstance(node, MibTable):
            hint = '# Table %s::%s\r\n' % (mibName, symName)
            if verboseFlag:
                sys.stderr.write('# Starting table %s::%s (%s)\r\n' % (mibName, symName, univ.ObjectIdentifier(oid)))
            continue
        elif isinstance(node, MibTableRow):
            rowIndices = {}
            suffix = ()
            rowHint = hint + '# Row %s::%s\r\n' % (mibName, symName)
            for impliedFlag, idxModName, idxSymName in node.getIndexNames():
                idxNode, = mibBuilder.importSymbols(idxModName, idxSymName)
                rowHint += '# Index %s::%s (type %s)\r\n' % (idxModName, idxSymName, idxNode.syntax.__class__.__name__)
                rowIndices[idxNode.name] = getValue(idxNode.syntax, verboseFlag and rowHint or '')
                suffix = suffix + node.getAsName(rowIndices[idxNode.name], impliedFlag)
            if not rowIndices:
                if verboseFlag:
                    sys.stderr.write('# WARNING: %s::%s table has no index!\r\n' % (mibName, symName))
            if rowOID is None:
                thisTableSize = 0
            rowOID = univ.ObjectIdentifier(oid)
            continue
        elif isinstance(node, MibTableColumn):
            oid = node.name
            if oid in rowIndices:
                val = rowIndices[oid]
            else:
                val = getValue(node.syntax,
                               verboseFlag and rowHint + '# Column %s::%s (type %s)\r\n' % (mibName, symName, node.syntax.__class__.__name__) or '')
        elif isinstance(node, MibScalar):
            hint = ''
            oid = node.name
            suffix = (0,)
            val = getValue(node.syntax, verboseFlag and '# Scalar %s::%s (type %s)\r\n' % (mibName, symName, node.syntax.__class__.__name__) or '')
        else:
            hint = ''
            continue

        output.append((oid + suffix, val))

    output.sort(key=lambda x: univ.ObjectIdentifier(x[0]))

    unique = set()

    for oid, val in output:
        if oid in unique:
            if verboseFlag:
                sys.stderr.write('# Dropping duplicate OID %s\r\n' % (univ.ObjectIdentifier(oid),))
        else:
            try:
                outputFile.write(
                    dataFileHandler.format(oid, val)
                )
            except SnmpsimError:
                sys.stderr.write('ERROR: %s\r\n' % (sys.exc_info()[1],))
            else:
                unique.add(oid)

    if verboseFlag:
        sys.stderr.write('# End of %s, %s OID(s) dumped\r\n' % (modName, len(unique)))

outputFile.flush()
outputFile.close()



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from pysnmp.proto import rfc1902
from pyasn1.type import univ
from pyasn1.compat.octets import octs2str
from snmpsim.grammar import abstract
from snmpsim import error


class SapGrammar(abstract.AbstractGrammar):
    tagMap = {
        'Counter': rfc1902.Counter32,
        'Gauge': rfc1902.Gauge32,
        'Integer': rfc1902.Integer32,
        'IpAddress': rfc1902.IpAddress,
        #        '<not implemented?>': univ.Null,
        'ObjectID': univ.ObjectIdentifier,
        'OctetString': rfc1902.OctetString,
        'TimeTicks': rfc1902.TimeTicks,
        'Counter64': rfc1902.Counter64
    }

    def __stringFilter(value):
        if value[:2] == '0x':
            value = [int(value[x:x + 2], 16) for x in range(2, len(value[2:]) + 2, 2)]
        return value

    filterMap = {
        'OctetString': __stringFilter
    }

    def parse(self, line):
        try:
            oid, tag, value = [x.strip() for x in octs2str(line).split(',', 2)]
        except:
            raise error.SnmpsimError('broken record <%s>' % line)
        else:
            if oid and tag:
                return oid, tag, self.filterMap.get(tag, lambda x: x)(value.strip())
            raise error.SnmpsimError('broken record <%s>' % line)



#!/usr/bin/env python
#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Simulator MIB to data file converter
#
import getopt
import sys
import os
import time
import socket
import struct
import bisect
import traceback

try:
    import pcap
except ImportError:
    pcap = None
from pyasn1.type import univ
from pyasn1.codec.ber import decoder
from pyasn1.error import PyAsn1Error
from pysnmp.proto import api, rfc1905
from pysnmp.smi import builder, rfc1902, view, compiler
from pysnmp.carrier.asynsock.dgram import udp
from pyasn1 import debug as pyasn1_debug
from pysnmp import debug as pysnmp_debug
from snmpsim.record import snmprec
from snmpsim import confdir, error, log

# Defaults
PROGRAM_NAME = 'pcap2dev'
verboseFlag = True
mibSources = []
defaultMibSources = ['http://mibs.snmplabs.com/asn1/@mib@']
startOID = univ.ObjectIdentifier('1.3.6')
stopOID = None
promiscuousMode = False
outputDir = '.'
transportIdOffset = 0
loggingMethod = ['stderr']
loggingLevel = None
variationModuleOptions = ""
variationModuleName = variationModule = None
listenInterface = captureFile = None
packetFilter = 'udp and src port 161'

endpoints = {}
contexts = {}

stats = {
    'UDP packets': 0,
    'IP packets': 0,
    'bad packets': 0,
    'empty packets': 0,
    'unknown L2 protocol': 0,
    'SNMP errors': 0,
    'SNMP exceptions': 0,
    'agents seen': 0,
    'contexts seen': 0,
    'snapshots taken': 0,
    'Response PDUs seen': 0,
    'OIDs seen': 0
}

helpMessage = """Usage: %s [--help]
    [--version]
    [--debug=<%s>]
    [--debug-asn1=<%s>]
    [--quiet]
    [--logging-method=<%s[:args]>]
    [--log-level=<%s>]
    [--mib-source=<url>]
    [--start-object=<MIB-NAME::[symbol-name]|OID>]
    [--stop-object=<MIB-NAME::[symbol-name]|OID>]
    [--output-dir=<directory>]
    [--transport-id-offset=<number>]
    [--capture-file=<filename.pcap>]
    [--listen-interface=<device>]
    [--promiscuous-mode]
    [--packet-filter=<ruleset>]
    [--variation-modules-dir=<dir>]
    [--variation-module=<module>]
    [--variation-module-options=<args>]""" % (
    sys.argv[0],
    '|'.join([x for x in getattr(pysnmp_debug, 'FLAG_MAP', getattr(pysnmp_debug, 'flagMap', ()))
              if x != 'mibview']),
    '|'.join([x for x in getattr(pyasn1_debug, 'FLAG_MAP', getattr(pyasn1_debug, 'flagMap', ()))]),
    '|'.join(log.METHODS_MAP),
    '|'.join(log.LEVELS_MAP)
)

try:
    opts, params = getopt.getopt(sys.argv[1:], 'hv', [
        'help', 'version', 'debug=', 'debug-snmp=', 'debug-asn1=',
        'quiet', 'logging-method=', 'log-level=', 'start-oid=', 'stop-oid=',
        'start-object=', 'stop-object=', 'mib-source=',
        'output-dir=', 'transport-id-offset=',
        'capture-file=', 'listen-interface=', 'promiscuous-mode',
        'packet-filter=',
        'variation-modules-dir=', 'variation-module=',
        'variation-module-options='
    ])

except Exception:
    sys.stderr.write(
        'ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
    sys.exit(-1)

if params:
    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
    sys.exit(-1)

for opt in opts:
    if opt[0] == '-h' or opt[0] == '--help':
        sys.stderr.write("""\
Synopsis:
  Snoops network traffic for SNMP responses, builds SNMP Simulator
  data files.
  Can read capture files or listen live network interface.
Documentation:
  http://snmplabs.com/snmpsim/
%s
""" % helpMessage)
        sys.exit(-1)
    if opt[0] == '-v' or opt[0] == '--version':
        import snmpsim
        import pysmi
        import pysnmp
        import pyasn1

        sys.stderr.write("""\
SNMP Simulator version %s, written by Ilya Etingof <etingof@gmail.com>
Using foundation libraries: pysmi %s, pysnmp %s, pyasn1 %s.
Python interpreter: %s
Software documentation and support at http://snmplabs.com/snmpsim
%s
""" % (snmpsim.__version__,
       hasattr(pysmi, '__version__') and pysmi.__version__ or 'unknown',
       hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown',
       hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown',
       sys.version, helpMessage))
        sys.exit(-1)
    elif opt[0] in ('--debug', '--debug-snmp'):
        pysnmp_debug.setLogger(pysnmp_debug.Debug(*opt[1].split(','), **dict(
            loggerName='pcap2dev.pysnmp')))
    elif opt[0] == '--debug-asn1':
        pyasn1_debug.setLogger(pyasn1_debug.Debug(*opt[1].split(','), **dict(
            loggerName='pcap2dev.pyasn1')))
    elif opt[0] == '--logging-method':
        loggingMethod = opt[1].split(':')
    elif opt[0] == '--log-level':
        loggingLevel = opt[1]
    if opt[0] == '--quiet':
        verboseFlag = False
    # obsolete begin
    elif opt[0] == '--start-oid':
        startOID = univ.ObjectIdentifier(opt[1])
    elif opt[0] == '--stop-oid':
        stopOID = univ.ObjectIdentifier(opt[1])
    # obsolete end
    if opt[0] == '--mib-source':
        mibSources.append(opt[1])
    if opt[0] == '--start-object':
        startOID = rfc1902.ObjectIdentity(*opt[1].split('::', 1))
    if opt[0] == '--stop-object':
        stopOID = rfc1902.ObjectIdentity(*opt[1].split('::', 1),
                                         **dict(last=True))
    elif opt[0] == '--output-dir':
        outputDir = opt[1]
    elif opt[0] == '--transport-id-offset':
        try:
            transportIdOffset = max(0, int(opt[1]))
        except:
            sys.stderr.write(
                'ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    elif opt[0] == '--listen-interface':
        listenInterface = opt[1]
    elif opt[0] == '--promiscuous-mode':
        promiscuousMode = True
    elif opt[0] == '--capture-file':
        captureFile = opt[1]
    elif opt[0] == '--packet-filter':
        packetFilter = opt[1]
    elif opt[0] == '--variation-modules-dir':
        confdir.variation.insert(0, opt[1])
    elif opt[0] == '--variation-module':
        variationModuleName = opt[1]
    elif opt[0] == '--variation-module-options':
        variationModuleOptions = opt[1]

if params:
    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (
    params, helpMessage))
    sys.exit(-1)

if not pcap:
    sys.stderr.write('ERROR: pylibpcap package is missing!\r\nGet it by running `pip install https://downloads.sourceforge.net/project/pylibpcap/pylibpcap/0.6.4/pylibpcap-0.6.4.tar.gz`\r\n%s\r\n' % helpMessage)
    sys.exit(-1)

try:
    log.setLogger(PROGRAM_NAME, *loggingMethod, force=True)

    if loggingLevel:
        log.setLevel(loggingLevel)

except error.SnmpsimError:
    sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
    sys.exit(1)

if isinstance(startOID, rfc1902.ObjectIdentity) or \
        isinstance(stopOID, rfc1902.ObjectIdentity):
    mibBuilder = builder.MibBuilder()

    mibViewController = view.MibViewController(mibBuilder)

    compiler.addMibCompiler(
        mibBuilder, sources=mibSources or defaultMibSources
    )
    if isinstance(startOID, rfc1902.ObjectIdentity):
        startOID.resolveWithMib(mibViewController)
    if isinstance(stopOID, rfc1902.ObjectIdentity):
        stopOID.resolveWithMib(mibViewController)

# Load variation module

if variationModuleName:
    for variationModulesDir in confdir.variation:
        log.info('Scanning "%s" directory for variation modules...' % variationModulesDir)
        if not os.path.exists(variationModulesDir):
            log.info('Directory "%s" does not exist' % variationModulesDir)
            continue

        mod = os.path.join(variationModulesDir, variationModuleName + '.py')
        if not os.path.exists(mod):
            log.info('Variation module "%s" not found' % mod)
            continue

        ctx = {'path': mod, 'moduleContext': {}}

        try:
            if sys.version_info[0] > 2:
                exec(compile(open(mod).read(), mod, 'exec'), ctx)
            else:
                execfile(mod, ctx)

        except Exception:
            log.error('Variation module "%s" execution failure: %s' % (mod, sys.exc_info()[1]))
            sys.exit(-1)

        else:
            variationModule = ctx
            log.info('Variation module "%s" loaded' % variationModuleName)
            break
    else:
        log.error('variation module "%s" not found' % variationModuleName)
        sys.exit(-1)

# Variation module initialization

if variationModule:
    log.info('Initializing variation module...')
    for x in ('init', 'record', 'shutdown'):
        if x not in variationModule:
            log.error('missing "%s" handler at variation module "%s"' % (x, variationModuleName))
            sys.exit(-1)
    try:
        variationModule['init'](options=variationModuleOptions,
                                mode='recording',
                                startOID=startOID,
                                stopOID=stopOID)

    except Exception:
        log.error('Variation module "%s" initialization FAILED: %s' % (variationModuleName, sys.exc_info()[1]))

    else:
        log.info('Variation module "%s" initialization OK' % variationModuleName)


# Data file builder

class SnmprecRecord(snmprec.SnmprecRecord):
    def formatValue(self, oid, value, **context):
        textOid, textTag, textValue = snmprec.SnmprecRecord.formatValue(
            self, oid, value
        )

        # invoke variation module
        if context['variationModule']:
            plainOid, plainTag, plainValue = snmprec.SnmprecRecord.formatValue(
                self, oid, value, nohex=True
            )
            if plainTag != textTag:
                context['hextag'], context['hexvalue'] = textTag, textValue
            else:
                textTag, textValue = plainTag, plainValue

            textOid, textTag, textValue = context['variationModule'][
                'record'](
                textOid, textTag, textValue, **context
            )

        elif 'stopFlag' in context and context['stopFlag']:
            raise error.NoDataNotification()

        return textOid, textTag, textValue


pcapObj = pcap.pcapObject()

if listenInterface:
    if verboseFlag:
        log.info('Listening on interface %s in %spromiscuous mode' % (listenInterface, promiscuousMode is False and 'non-' or ''))
    try:
        pcapObj.open_live(listenInterface, 65536, promiscuousMode, 1000)
    except Exception:
        log.error('Error opening interface %s for snooping: %s' % (listenInterface, sys.exc_info()[1]))
        sys.exit(-1)
elif captureFile:
    if verboseFlag:
        log.info('Opening capture file %s' % captureFile)
    try:
        pcapObj.open_offline(captureFile)
    except Exception:
        log.error('Error opening capture file %s for reading: %s' % (captureFile, sys.exc_info()[1]))
        sys.exit(-1)
else:
    sys.stderr.write('ERROR: no capture file or live interface specified\r\n%s\r\n' % helpMessage)
    sys.exit(-1)

if packetFilter:
    if verboseFlag:
        log.info('Applying packet filter \"%s\"' % packetFilter)
    pcapObj.setfilter(packetFilter, 0, 0)

if verboseFlag:
    log.info('Processing records from %s till %s' % (startOID or 'the beginning', stopOID or 'the end'))


def parsePacket(s):
    d = {}

    # http://www.tcpdump.org/linktypes.html
    llHeaders = {
        0: 4,
        1: 14,
        108: 4,
        228: 0
    }

    if pcapObj.datalink() in llHeaders:
        s = s[llHeaders[pcapObj.datalink()]:]
    else:
        stats['unknown L2 protocol'] += 1

    d['version'] = (ord(s[0]) & 0xf0) >> 4
    d['header_len'] = ord(s[0]) & 0x0f
    d['tos'] = ord(s[1])
    d['total_len'] = socket.ntohs(struct.unpack('H', s[2:4])[0])
    d['id'] = socket.ntohs(struct.unpack('H', s[4:6])[0])
    d['flags'] = (ord(s[6]) & 0xe0) >> 5
    d['fragment_offset'] = socket.ntohs(struct.unpack('H', s[6:8])[0] & 0x1f)
    d['ttl'] = ord(s[8])
    d['protocol'] = ord(s[9])
    d['checksum'] = socket.ntohs(struct.unpack('H', s[10:12])[0])
    d['source_address'] = pcap.ntoa(struct.unpack('i', s[12:16])[0])
    d['destination_address'] = pcap.ntoa(struct.unpack('i', s[16:20])[0])
    if d['header_len'] > 5:
        d['options'] = s[20:4 * (d['header_len'] - 5)]
    else:
        d['options'] = None
    s = s[4 * d['header_len']:]
    if d['protocol'] == 17:
        d['source_port'] = socket.ntohs(struct.unpack('H', s[0:2])[0])
        d['destination_port'] = socket.ntohs(struct.unpack('H', s[2:4])[0])
        s = s[8:]
        stats['UDP packets'] += 1
    d['data'] = s
    stats['IP packets'] += 1
    return d


def handleSnmpMessage(d, t, private={}):
    msgVer = api.decodeMessageVersion(d['data'])
    if msgVer in api.protoModules:
        pMod = api.protoModules[msgVer]
    else:
        stats['bad packets'] += 1
        return
    try:
        rspMsg, wholeMsg = decoder.decode(
            d['data'], asn1Spec=pMod.Message(),
        )
    except PyAsn1Error:
        stats['bad packets'] += 1
        return
    if rspMsg['data'].getName() == 'response':
        rspPDU = pMod.apiMessage.getPDU(rspMsg)
        errorStatus = pMod.apiPDU.getErrorStatus(rspPDU)
        if errorStatus:
            stats['SNMP errors'] += 1
        else:
            endpoint = d['source_address'], d['source_port']
            if endpoint not in endpoints:
                endpoints[endpoint] = udp.domainName + (transportIdOffset + len(endpoints),)
                stats['agents seen'] += 1
            context = '%s/%s' % (pMod.ObjectIdentifier(endpoints[endpoint]),
                                 pMod.apiMessage.getCommunity(rspMsg))
            if context not in contexts:
                contexts[context] = {}
                stats['contexts seen'] += 1
            context = '%s/%s' % (pMod.ObjectIdentifier(endpoints[endpoint]),
                                 pMod.apiMessage.getCommunity(rspMsg))

            stats['Response PDUs seen'] += 1

            if 'basetime' not in private:
                private['basetime'] = t

            for oid, value in pMod.apiPDU.getVarBinds(rspPDU):
                if oid < startOID:
                    continue
                if stopOID and oid >= stopOID:
                    continue
                if oid in contexts[context]:
                    if value != contexts[context][oid]:
                        stats['snapshots taken'] += 1
                else:
                    contexts[context][oid] = [], []
                contexts[context][oid][0].append(t - private['basetime'])
                contexts[context][oid][1].append(value)
                stats['OIDs seen'] += 1


def handlePacket(pktlen, data, timestamp):
    if not data:
        stats['empty packets'] += 1
        return
    else:
        handleSnmpMessage(parsePacket(data), timestamp)


exc_info = None

try:
    if listenInterface:
        log.info(
            'Listening on interface "%s", kill me when you are done.' % listenInterface)
        while True:
            pcapObj.dispatch(1, handlePacket)
    elif captureFile:
        log.info('Processing capture file "%s"....' % captureFile)
        args = pcapObj.next()
        while args:
            handlePacket(*args)
            args = pcapObj.next()

except (TypeError, KeyboardInterrupt):
    log.info('Shutting down process...')

except Exception:
    exc_info = sys.exc_info()

dataFileHandler = SnmprecRecord()

for context in contexts:
    filename = os.path.join(outputDir,
                            context + os.path.extsep + SnmprecRecord.ext)
    if verboseFlag:
        log.info('Creating simulation context %s at %s' % (context, filename))
    try:
        os.mkdir(os.path.dirname(filename))
    except OSError:
        pass
    try:
        outputFile = open(filename, 'wb')
    except IOError:
        log.error('writing %s: %s' % (filename, sys.exc_info()[1]))
        sys.exit(-1)

    count = total = iteration = 0
    timeOffset = 0
    reqTime = time.time()
    oids = list(contexts[context].keys())
    oids.sort()
    oids.append(oids[-1])  # duplicate last OID to trigger stopFlag
    while True:
        for oid in oids:
            timeline, values = contexts[context][oid]
            value = values[
                min(len(values) - 1,
                    bisect.bisect_left(timeline, timeOffset))
            ]
            if value.tagSet in (rfc1905.NoSuchObject.tagSet,
                                rfc1905.NoSuchInstance.tagSet,
                                rfc1905.EndOfMibView.tagSet):
                stats['SNMP exceptions'] += 1
                continue

            # remove value enumeration

            if value.tagSet == rfc1902.Integer32.tagSet:
                value = rfc1902.Integer32(value)

            if value.tagSet == rfc1902.Unsigned32.tagSet:
                value = rfc1902.Unsigned32(value)

            if value.tagSet == rfc1902.Bits.tagSet:
                value = rfc1902.OctetString(value)

            # Build .snmprec record

            ctx = {
                'origOid': oid,
                'origValue': value,
                'count': count,
                'total': total,
                'iteration': iteration,
                'reqTime': reqTime,
                'startOID': startOID,
                'stopOID': stopOID,
                'stopFlag': oids.index(oid) == len(oids) - 1,
                'variationModule': variationModule
            }

            try:
                line = dataFileHandler.format(oid, value, **ctx)
            except error.MoreDataNotification:
                count = 0
                iteration += 1

                moreDataNotification = sys.exc_info()[1]
                if 'period' in moreDataNotification:
                    timeOffset += moreDataNotification['period']
                    log.info(
                        '%s OIDs dumped, advancing time window to %.2f sec(s)...' % (total, timeOffset))
                break

            except error.NoDataNotification:
                pass
            except error.SnmpsimError:
                log.error((sys.exc_info()[1],))
                continue
            else:
                outputFile.write(line)

                count += 1
                total += 1

        else:
            break

    outputFile.flush()
    outputFile.close()

if variationModule:
    log.info('Shutting down variation module "%s"...' % variationModuleName)
    try:
        variationModule['shutdown'](options=variationModuleOptions,
                                    mode='recording')
    except Exception:
        log.error('Variation module "%s" shutdown FAILED: %s' % (variationModuleName, sys.exc_info()[1]))

    else:
        log.info('Variation module "%s" shutdown OK' % variationModuleName)

log.info("""PCap statistics:
    packets snooped: %s
    packets dropped: %s
    packets dropped: by interface %s""" % pcapObj.stats())
log.info("""SNMP statistics:
    %s""" % '    '.join(['%s: %s\r\n' % kv for kv in stats.items()]))

if exc_info:
    for line in traceback.format_exception(*exc_info):
        log.error(line.replace('\n', ';'))



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from string import digits, ascii_letters
from pysnmp.proto import rfc1902, rfc1905
from pyasn1.compat.octets import octs2str, str2octs, octs2ints
from pyasn1.type import univ
from snmpsim.grammar.abstract import AbstractGrammar
from snmpsim import error


class SnmprecGrammar(AbstractGrammar):
    alnums = set(octs2ints(str2octs(ascii_letters + digits)))
    tagMap = {}
    for t in (rfc1902.Gauge32,
              rfc1902.Integer32,
              rfc1902.IpAddress,
              univ.Null,
              univ.ObjectIdentifier,
              rfc1902.OctetString,
              rfc1902.TimeTicks,
              rfc1902.Opaque,
              rfc1902.Counter32,
              rfc1902.Counter64,
              rfc1905.NoSuchObject,
              rfc1905.NoSuchInstance,
              rfc1905.EndOfMibView):
        tagMap[str(sum([x for x in t.tagSet[0]]))] = t

    def build(self, oid, tag, val):
        if oid and tag:
            return str2octs('%s|%s|%s\n' % (oid, tag, val))
        raise error.SnmpsimError('empty OID/tag <%s/%s>' % (oid, tag))

    def parse(self, line):
        try:
            oid, tag, value = octs2str(line).strip().split('|', 2)
        except:
            raise error.SnmpsimError('broken record <%s>' % line)
        else:
            if oid and tag:
                return oid, tag, value
            raise error.SnmpsimError('broken record <%s>' % line)

    # helper functions

    def getTagByType(self, value):
        for tag, typ in self.tagMap.items():
            if typ.tagSet[0] == value.tagSet[0]:
                return tag
        raise Exception('error: unknown type of %s' % (value,))

    def hexifyValue(self, value):
        if value.tagSet in (univ.OctetString.tagSet,
                            rfc1902.Opaque.tagSet,
                            rfc1902.IpAddress.tagSet):
            nval = value.asNumbers()
            for x in nval:
                if (value.tagSet == rfc1902.IpAddress.tagSet or
                            x not in self.alnums):
                    return ''.join(['%.2x' % x for x in nval])



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# Like string.split but first tries to use composite separator as an
# escaping aid


def split(val, sep):
    for x in (3, 2, 1):
        if val.find(sep * x) != -1:
            return val.split(sep * x)
    return [val]



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Simulator, http://snmplabs.com/snmpsim
# Managed value variaton module
# Simulate a writable Agent
#
import shelve
from pysnmp.smi import error
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim.mltsplit import split
from snmpsim import log

errorTypes = {
    'generror': error.GenError,
    'noaccess': error.NoAccessError,
    'wrongtype': error.WrongTypeError,
    'wrongvalue': error.WrongValueError,
    'nocreation': error.NoCreationError,
    'inconsistentvalue': error.InconsistentValueError,
    'resourceunavailable': error.ResourceUnavailableError,
    'commitfailed': error.CommitFailedError,
    'undofailed': error.UndoFailedError,
    'authorizationerror': error.AuthorizationError,
    'notwritable': error.NotWritableError,
    'inconsistentname': error.InconsistentNameError,
    'nosuchobject': error.NoSuchObjectError,
    'nosuchinstance': error.NoSuchInstanceError,
    'endofmib': error.EndOfMibViewError
}


def init(**context):
    moduleContext['settings'] = {}
    if context['options']:
        moduleContext['settings'].update(
            dict([split(x, ':') for x in split(context['options'], ',')])
        )
    if 'file' in moduleContext['settings']:
        moduleContext['cache'] = shelve.open(moduleContext['settings']['file'])
    else:
        moduleContext['cache'] = {}


def variate(oid, tag, value, **context):
    if not context['nextFlag'] and not context['exactMatch']:
        return context['origOid'], tag, context['errorStatus']

    if 'settings' not in recordContext:
        recordContext['settings'] = dict([split(x, '=') for x in split(value, ',')])

        if 'vlist' in recordContext['settings']:
            vlist = {}
            recordContext['settings']['vlist'] = split(recordContext['settings']['vlist'], ':')
            while recordContext['settings']['vlist']:
                o, v, e = recordContext['settings']['vlist'][:3]
                recordContext['settings']['vlist'] = recordContext['settings']['vlist'][3:]
                typeTag, _ = SnmprecRecord.unpackTag(tag)
                v = SnmprecGrammar.tagMap[typeTag](v)
                if o not in vlist:
                    vlist[o] = {}
                if o == 'eq':
                    vlist[o][v] = e
                elif o in ('lt', 'gt'):
                    vlist[o] = v, e
                else:
                    log.msg('writecache: bad vlist syntax: %s' % recordContext['settings']['vlist'])

            recordContext['settings']['vlist'] = vlist

        if 'status' in recordContext['settings']:
            recordContext['settings']['status'] = recordContext['settings']['status'].lower()

    if oid not in moduleContext:
        moduleContext[oid] = {}
        typeTag, _ = SnmprecRecord.unpackTag(tag)
        moduleContext[oid]['type'] = SnmprecGrammar.tagMap[typeTag]()

    textOid = str(oid)

    if context['setFlag']:
        if 'vlist' in recordContext['settings']:
            if ('eq' in recordContext['settings']['vlist'] and
                    context['origValue'] in recordContext['settings']['vlist']['eq']):
                e = recordContext['settings']['vlist']['eq'][context['origValue']]
            elif ('lt' in recordContext['settings']['vlist'] and
                    context['origValue'] < recordContext['settings']['vlist']['lt'][0]):
                e = recordContext['settings']['vlist']['lt'][1]
            elif ('gt' in recordContext['settings']['vlist'] and
                    context['origValue'] > recordContext['settings']['vlist']['gt'][0]):
                e = recordContext['settings']['vlist']['gt'][1]
            else:
                e = None

            if e in errorTypes:
                raise errorTypes[e](
                    name=oid, idx=max(0, context['varsTotal'] - context['varsRemaining'] - 1)
                )

        if moduleContext[oid]['type'].isSameTypeWith(context['origValue']):
            moduleContext['cache'][textOid] = context['origValue']
        else:
            return context['origOid'], tag, context['errorStatus']

    if 'status' in recordContext['settings']:
        if ('op' not in recordContext['settings'] or
                recordContext['settings']['op'] == 'any' or
                recordContext['settings']['op'] == 'set' and context['setFlag'] or
                recordContext['settings']['op'] == 'get' and not context['setFlag']):
            e = recordContext['settings']['status']
            if e in errorTypes:
                raise errorTypes[e](
                    name=oid, idx=max(0, context['varsTotal'] - context['varsRemaining'] - 1)
                )

    if textOid in moduleContext['cache']:
        return oid, tag, moduleContext['cache'][textOid]
    elif 'hexvalue' in recordContext['settings']:
        return oid, tag, moduleContext[oid]['type'].clone(hexValue=recordContext['settings']['hexvalue'])
    elif 'value' in recordContext['settings']:
        return oid, tag, moduleContext[oid]['type'].clone(recordContext['settings']['value'])
    else:
        return oid, tag, context['errorStatus']


def shutdown(**context):
    if 'file' in moduleContext['settings']:
        moduleContext['cache'].close()



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
from snmpsim.grammar import dump


class MvcGrammar(dump.DumpGrammar): pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# Managed value variation module
# Simulate a numeric value
# Valid values in module options are:
#   2  - Integer
#   65 - Counter32
#   66 - Gauge32
#   67 - TimeTicks
#   70 - Counter64
import math
import time
import random
from pysnmp.proto import rfc1902
from snmpsim.mltsplit import split
from snmpsim import log
from snmpsim import error

tboot = time.time()


def init(**context):
    if context['mode'] == 'variating':
        random.seed()
    if context['mode'] == 'recording':
        moduleContext['settings'] = {}
        if context['options']:
            for k, v in [split(x, ':') for x in split(context['options'], ',')]:
                if k == 'addon':
                    if k in moduleContext['settings']:
                        moduleContext['settings'][k].append(v)
                    else:
                        moduleContext['settings'][k] = [v]
                else:
                    moduleContext['settings'][k] = v
            if 'iterations' in moduleContext['settings']:
                moduleContext['settings']['iterations'] = int(moduleContext['settings']['iterations'])
                if moduleContext['settings']['iterations']:
                    moduleContext['settings']['iterations'] = 1  # no reason for more
            if 'period' in moduleContext['settings']:
                moduleContext['settings']['period'] = float(moduleContext['settings']['period'])
            else:
                moduleContext['settings']['period'] = 10.0

            if 'taglist' not in moduleContext['settings']:
                moduleContext['settings']['taglist'] = '2-65-66-67-70'


def variate(oid, tag, value, **context):
    if not context['nextFlag'] and not context['exactMatch']:
        return context['origOid'], tag, context['errorStatus']
    if context['setFlag']:
        return context['origOid'], tag, context['errorStatus']

    if 'settings' not in recordContext:
        recordContext['settings'] = dict([split(x, '=') for x in split(value, ',')])
        for k in recordContext['settings']:
            if k != 'function':
                recordContext['settings'][k] = float(recordContext['settings'][k])
        if 'min' not in recordContext['settings']:
            recordContext['settings']['min'] = 0
        if 'max' not in recordContext['settings']:
            if tag == '70':
                recordContext['settings']['max'] = 0xffffffffffffffff
            else:
                recordContext['settings']['max'] = 0xffffffff
        if 'rate' not in recordContext['settings']:
            recordContext['settings']['rate'] = 1
        if 'function' in recordContext['settings']:
            f = split(recordContext['settings']['function'], '%')
            recordContext['settings']['function'] = getattr(math, f[0]), f[1:]
        else:
            recordContext['settings']['function'] = lambda x: x, ()

    vold, told = recordContext['settings'].get('initial', recordContext['settings']['min']), tboot

    if 'cumulative' in recordContext['settings']:
        if 'value' not in recordContext:
            recordContext['value'] = vold, told
        vold, told = recordContext['value']

    tnow = time.time()

    if 'atime' in recordContext['settings']:
        t = tnow
    else:
        t = tnow - tboot

    f, args = recordContext['settings']['function']

    _args = []
    if args:
        for x in args:
            if x == '<time>':
                _args.append(t * recordContext['settings']['rate'])
            else:
                _args.append(float(x))
    else:
        _args.append(t * recordContext['settings']['rate'])

    v = f(*_args)

    if 'scale' in recordContext['settings']:
        v *= recordContext['settings']['scale']

    if 'offset' in recordContext['settings']:
        if 'cumulative' in recordContext['settings']:
            v += recordContext['settings']['offset'] * (tnow - told) * recordContext['settings']['rate']
        else:
            v += recordContext['settings']['offset']

    if 'deviation' in recordContext['settings'] and recordContext['settings']['deviation']:
        v += random.randrange(-recordContext['settings']['deviation'], recordContext['settings']['deviation'])

    if 'cumulative' in recordContext['settings']:
        v = max(v, 0)

    v += vold

    if v < recordContext['settings']['min']:
        v = recordContext['settings']['min']
    elif v > recordContext['settings']['max']:
        if 'wrap' in recordContext['settings']:
            v %= recordContext['settings']['max']
            v += recordContext['settings']['min']
        else:
            v = recordContext['settings']['max']

    if 'cumulative' in recordContext['settings']:
        recordContext['value'] = v, tnow

    return oid, tag, v


def record(oid, tag, value, **context):
    if 'started' not in moduleContext:
        moduleContext['started'] = time.time()

    if 'iterations' not in moduleContext:
        moduleContext['iterations'] = min(1, moduleContext['settings'].get('iterations', 0))

    # single-run recording

    if 'iterations' not in moduleContext['settings'] or not moduleContext['settings']['iterations']:
        if context['origValue'].tagSet not in (
                rfc1902.Counter32.tagSet,
                rfc1902.Counter64.tagSet,
                rfc1902.TimeTicks.tagSet,
                rfc1902.Gauge32.tagSet,
                rfc1902.Integer.tagSet):
            if 'hextag' in context:
                tag = context['hextag']
            if 'hexvalue' in context:
                value = context['hexvalue']
            return oid, tag, value

        if ('taglist' not in moduleContext['settings'] or
                tag not in moduleContext['settings']['taglist']):
            return oid, tag, value

        value = 'initial=%s' % value

        if context['origValue'].tagSet == rfc1902.TimeTicks.tagSet:
            value += ',rate=100'
        elif context['origValue'].tagSet == rfc1902.Integer.tagSet:
            value += ',rate=0'

        return oid, tag + ':numeric', value

    # multiple-iteration recording

    if oid not in moduleContext:
        settings = {
            'initial': value
        }
        if context['origValue'].tagSet == rfc1902.TimeTicks.tagSet:
            settings['rate'] = 100
        elif context['origValue'].tagSet == rfc1902.Integer.tagSet:
            settings['rate'] = 0  # may be constants
        if 'addon' in moduleContext['settings']:
            settings.update(
                dict([split(x, '=') for x in moduleContext['settings']['addon']])
            )

        moduleContext[oid] = {}

        moduleContext[oid]['settings'] = settings

    if moduleContext['iterations']:
        if context['stopFlag']:  # switching to final iteration
            log.msg('numeric: %s iterations remaining' % moduleContext['settings']['iterations'])
            moduleContext['iterations'] -= 1
            moduleContext['started'] = time.time()
            wait = max(0, float(moduleContext['settings']['period']) - (time.time() - moduleContext['started']))
            raise error.MoreDataNotification(period=wait)
        else:  # storing values on first iteration
            moduleContext[oid]['time'] = time.time()
            moduleContext[oid]['value'] = context['origValue']
            if 'hexvalue' in moduleContext[oid]:
                moduleContext[oid]['hexvalue'] = context['hexvalue']
            if 'hextag' in moduleContext[oid]:
                moduleContext[oid]['hextag'] = context['hextag']
            raise error.NoDataNotification()
    else:
        if context['stopFlag']:
            raise error.NoDataNotification()

        if 'value' in moduleContext[oid]:
            if context['origValue'].tagSet not in (
                    rfc1902.Counter32.tagSet,
                    rfc1902.Counter64.tagSet,
                    rfc1902.TimeTicks.tagSet,
                    rfc1902.Gauge32.tagSet,
                    rfc1902.Integer.tagSet):
                if 'hextag' in moduleContext[oid]:
                    tag = moduleContext[oid]['hextag']
                if 'hexvalue' in moduleContext[oid]:
                    value = moduleContext[oid]['hexvalue']
                return oid, tag, value

            if tag not in moduleContext['settings']['taglist']:
                return oid, tag, moduleContext[oid]['value']

            moduleContext[oid]['settings']['rate'] = (int(context['origValue']) - int(moduleContext[oid]['value'])) / (time.time() - moduleContext[oid]['time'])

            tag += ':numeric'
            value = ','.join(
                ['%s=%s' % (k, v) for k, v in moduleContext[oid]['settings'].items()]
            )
            return oid, tag, value
        else:
            raise error.NoDataNotification()


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
import os
import sys
import tempfile

if sys.platform[:3] == 'win':
    variation = [
        os.path.join(os.environ['HOMEPATH'], 'SNMP Simulator', 'Variation'),
        os.path.join(os.environ['APPDATA'], 'SNMP Simulator', 'Variation'),
        os.path.join(os.environ['PROGRAMFILES'], 'SNMP Simulator', 'Variation'),
        os.path.join(os.path.split(__file__)[0], 'variation')
    ]
    data = [
        os.path.join(os.environ['HOMEPATH'], 'SNMP Simulator', 'Data'),
        os.path.join(os.environ['APPDATA'], 'SNMP Simulator', 'Data'),
        os.path.join(os.environ['PROGRAMFILES'], 'SNMP Simulator', 'Data'),
        os.path.join(os.path.split(__file__)[0], 'data')
    ]
elif sys.platform == 'darwin':
    variation = [
        os.path.join(os.environ['HOME'], '.snmpsim', 'variation'),
        os.path.join('/', 'usr', 'local', 'share', 'snmpsim', 'variation'),
        os.path.join(sys.prefix, 'snmpsim', 'variation'),
        os.path.join(sys.prefix, 'share', 'snmpsim', 'variation'),
        os.path.join(os.path.split(__file__)[0], 'variation')
    ]
    data = [
        os.path.join(os.environ['HOME'], '.snmpsim', 'data'),
        os.path.join('/', 'usr', 'local', 'share', 'snmpsim', 'data'),
        os.path.join(sys.prefix, 'snmpsim', 'data'),
        os.path.join(sys.prefix, 'share', 'snmpsim', 'data'),
        os.path.join(os.path.split(__file__)[0], 'data')
    ]
else:
    variation = [
        os.path.join(os.environ['HOME'], '.snmpsim', 'variation'),
        os.path.join(sys.prefix, 'snmpsim', 'variation'),
        os.path.join(sys.prefix, 'share', 'snmpsim', 'variation'),
        os.path.join(os.path.split(__file__)[0], 'variation')
    ]
    data = [
        os.path.join(os.environ['HOME'], '.snmpsim', 'data'),
        os.path.join(sys.prefix, 'snmpsim', 'data'),
        os.path.join(sys.prefix, 'share', 'snmpsim', 'data'),
        os.path.join(os.path.split(__file__)[0], 'data')
    ]

cache = os.path.join(tempfile.gettempdir(), 'snmpsim')



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
import sys
from pyasn1.type import univ
from pyasn1.error import PyAsn1Error
from snmpsim.grammar import dump
from snmpsim.error import SnmpsimError
from snmpsim.record import abstract


class DumpRecord(abstract.AbstractRecord):
    grammar = dump.DumpGrammar()
    ext = 'dump'

    def evaluateOid(self, oid):
        return univ.ObjectIdentifier(oid)

    def evaluateValue(self, oid, tag, value, **context):
        try:
            value = self.grammar.tagMap[tag](value)
        except:
            raise SnmpsimError('value evaluation error for tag %r, value %r' % (tag, value))

        # not all callers supply the context - just ignore it
        try:
            if (not context['nextFlag'] and
                not context['exactMatch'] or
                    context['setFlag']):
                return context['origOid'], tag, context['errorStatus']

        except KeyError:
            pass

        return oid, tag, value

    def evaluate(self, line, **context):
        oid, tag, value = self.grammar.parse(line)
        oid = self.evaluateOid(oid)
        if context.get('oidOnly'):
            value = None
        else:
            try:
                oid, tag, value = self.evaluateValue(oid, tag, value, **context)
            except PyAsn1Error:
                raise SnmpsimError('value evaluation for %s = %r failed: %s\r\n' % (oid, value, sys.exc_info()[1]))
        return oid, value

    def formatOid(self, oid):
        return univ.ObjectIdentifier(oid).prettyPrint()

    def formatValue(self, oid, value, **context):
        return self.formatOid(oid), self.grammar.getTagByType(value), str(value)

    def format(self, oid, value, **context):
        textOid, textTag, textValue = self.formatValue(oid, value, **context)
        return self.grammar.build(
            textOid, textTag, textValue
        )



#!/usr/bin/env python
#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator
#
import os
import stat
import sys
import getopt
import traceback
if sys.version_info[0] < 3 and sys.version_info[1] < 5:
    from md5 import md5
else:
    from hashlib import md5
from pyasn1.type import univ
from pyasn1.codec.ber import encoder, decoder
from pyasn1.compat.octets import str2octs
from pyasn1.error import PyAsn1Error
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asynsock.dgram import udp
try:
    from pysnmp.carrier.asynsock.dgram import udp6
except ImportError:
    udp6 = None
try:
    from pysnmp.carrier.asynsock.dgram import unix
except ImportError:
    unix = None
from pysnmp.carrier.asynsock.dispatch import AsynsockDispatcher
from pysnmp.smi import exval, indices
from pysnmp.smi.error import MibOperationError
from pysnmp.proto import rfc1902, rfc1905, api
from pysnmp import error
from pyasn1 import debug as pyasn1_debug
from pysnmp import debug as pysnmp_debug
from snmpsim.error import SnmpsimError, NoDataNotification
from snmpsim import confdir, log, daemon
from snmpsim.record import dump, mvc, sap, walk, snmprec
from snmpsim.record.search.file import searchRecordByOid, getRecord
from snmpsim.record.search.database import RecordIndex

# Settings
PROGRAM_NAME = 'snmpsimd'
forceIndexBuild = False
validateData = False
maxVarBinds = 64
transportIdOffset = 0
v2cArch = False
v3Only = False
pidFile = '/var/run/snmpsim/snmpsimd.pid'
foregroundFlag = True
procUser = procGroup = None
loggingMethod = ['stderr']
loggingLevel = None
variationModulesOptions = {}
variationModules = {}

authProtocols = {
    'MD5': config.usmHMACMD5AuthProtocol,
    'SHA': config.usmHMACSHAAuthProtocol,
    'SHA224': config.usmHMAC128SHA224AuthProtocol,
    'SHA256': config.usmHMAC192SHA256AuthProtocol,
    'SHA384': config.usmHMAC256SHA384AuthProtocol,
    'SHA512': config.usmHMAC384SHA512AuthProtocol,
    'NONE': config.usmNoAuthProtocol
}

privProtocols = {
  'DES': config.usmDESPrivProtocol,
  '3DES': config.usm3DESEDEPrivProtocol,
  'AES': config.usmAesCfb128Protocol,
  'AES128': config.usmAesCfb128Protocol,
  'AES192': config.usmAesCfb192Protocol,
  'AES192BLMT': config.usmAesBlumenthalCfb192Protocol,
  'AES256': config.usmAesCfb256Protocol,
  'AES256BLMT': config.usmAesBlumenthalCfb256Protocol,
  'NONE': config.usmNoPrivProtocol
}


# Transport endpoints collection

class TransportEndpointsBase:
    def __init__(self):
        self.__endpoint = None

    def add(self, addr):
        self.__endpoint = self._addEndpoint(addr)
        return self

    def _addEndpoint(self, addr):
        raise NotImplementedError()

    def __len__(self):
        return len(self.__endpoint)

    def __getitem__(self, i):
        return self.__endpoint[i]


class IPv4TransportEndpoints(TransportEndpointsBase):
    def _addEndpoint(self, addr):
        f = lambda h, p=161: (h, int(p))
        try:
            h, p = f(*addr.split(':'))
        except:
            raise SnmpsimError('improper IPv4/UDP endpoint %s' % addr)
        return udp.UdpTransport().openServerMode((h, p)), addr


class IPv6TransportEndpoints(TransportEndpointsBase):
    def _addEndpoint(self, addr):
        if not udp6:
            raise SnmpsimError('This system does not support UDP/IP6')
        if addr.find(']:') != -1 and addr[0] == '[':
            h, p = addr.split(']:')
            try:
                h, p = h[1:], int(p)
            except:
                raise SnmpsimError('improper IPv6/UDP endpoint %s' % addr)
        elif addr[0] == '[' and addr[-1] == ']':
            h, p = addr[1:-1], 161
        else:
            h, p = addr, 161
        return udp6.Udp6Transport().openServerMode((h, p)), addr


class UnixTransportEndpoints(TransportEndpointsBase):
    def _addEndpoint(self, addr):
        if not unix:
            raise SnmpsimError('This system does not support UNIX domain sockets')
        return unix.UnixTransport().openServerMode(addr), addr


# Extended snmprec record handler

class SnmprecRecord(snmprec.SnmprecRecord):
    def evaluateValue(self, oid, tag, value, **context):
        # Variation module reference
        if ':' in tag:
            modName, tag = tag[tag.index(':')+1:], tag[:tag.index(':')]
        else:
            modName = None

        if modName:
            if ('variationModules' in context and
                    modName in context['variationModules']):
                if 'dataValidation' in context:
                    return oid, tag, univ.Null
                else:
                    if context['setFlag']:
                        hexvalue = self.grammar.hexifyValue(context['origValue'])
                        if hexvalue is not None:
                            context['hexvalue'] = hexvalue
                            context['hextag'] = self.grammar.getTagByType(context['origValue']) + 'x'

                    # prepare agent and record contexts on first reference
                    (variationModule,
                     agentContexts,
                     recordContexts) = context['variationModules'][modName]
                    if context['dataFile'] not in agentContexts:
                        agentContexts[context['dataFile']] = {}
                    if context['dataFile'] not in recordContexts:
                        recordContexts[context['dataFile']] = {}
                    variationModule['agentContext'] = agentContexts[context['dataFile']]
                    recordContexts = recordContexts[context['dataFile']]
                    if oid not in recordContexts:
                        recordContexts[oid] = {}

                    variationModule['recordContext'] = recordContexts[oid]

                    # invoke variation module
                    oid, tag, value = variationModule['variate'](oid, tag, value, **context)
            else:
                raise SnmpsimError('Variation module "%s" referenced but not loaded\r\n' % modName)

        if not modName:
            if 'dataValidation' in context:
                snmprec.SnmprecRecord.evaluateValue(
                    self, oid, tag, value, **context
                )

            if (not context['nextFlag'] and not context['exactMatch'] or
                    context['setFlag']):
                return context['origOid'], tag, context['errorStatus']

        if not hasattr(value, 'tagSet'):  # not already a pyasn1 object
            return snmprec.SnmprecRecord.evaluateValue(
                       self, oid, tag, value, **context
                   )

        return oid, tag, value

    def evaluate(self, line, **context):
        oid, tag, value = self.grammar.parse(line)
        oid = self.evaluateOid(oid)
        if context.get('oidOnly'):
            value = None
        else:
            try:
                oid, tag, value = self.evaluateValue(oid, tag, value, **context)
            except NoDataNotification:
                raise
            except MibOperationError:
                raise
            except PyAsn1Error:
                raise SnmpsimError('value evaluation for %s = %r failed: %s\r\n' % (oid, value, sys.exc_info()[1]))
        return oid, value

recordSet = {
    dump.DumpRecord.ext: dump.DumpRecord(),
    mvc.MvcRecord.ext: mvc.MvcRecord(),
    sap.SapRecord.ext: sap.SapRecord(),
    walk.WalkRecord.ext: walk.WalkRecord(),
    SnmprecRecord.ext: SnmprecRecord()
}


class AbstractLayout:
    layout = '?'


# Data text file and OID index

class DataFile(AbstractLayout):
    layout = 'text'
    openedQueue = []
    maxQueueEntries = 31  # max number of open text and index files

    def __init__(self, textFile, textParser):
        self.__recordIndex = RecordIndex(textFile, textParser)
        self.__textParser = textParser
        self.__textFile = os.path.abspath(textFile)
        
    def indexText(self, forceIndexBuild=False):
        self.__recordIndex.create(forceIndexBuild, validateData)
        return self

    def close(self):
        self.__recordIndex.close()
    
    def getHandles(self):
        if not self.__recordIndex.isOpen():
            if len(DataFile.openedQueue) > self.maxQueueEntries:
                log.info('Closing %s' % self)
                DataFile.openedQueue[0].close()
                del DataFile.openedQueue[0]

            DataFile.openedQueue.append(self)

            log.info('Opening %s' % self)

        return self.__recordIndex.getHandles()

    def processVarBinds(self, varBinds, **context):
        rspVarBinds = []

        if context.get('nextFlag'):
            errorStatus = exval.endOfMib
        else:
            errorStatus = exval.noSuchInstance

        try:
            text, db = self.getHandles()
        except SnmpsimError:
            log.error('Problem with data file or its index: %s' % sys.exc_info()[1])
            return [(vb[0], errorStatus) for vb in varBinds]

        varsRemaining = varsTotal = len(varBinds)

        log.info('Request var-binds: %s, flags: %s, %s' % (', '.join(['%s=<%s>' % (vb[0], vb[1].prettyPrint()) for vb in varBinds]), context.get('nextFlag') and 'NEXT' or 'EXACT', context.get('setFlag') and 'SET' or 'GET'))

        for oid, val in varBinds:
            textOid = str(univ.OctetString('.'.join(['%s' % x for x in oid])))

            try:
                line = self.__recordIndex.lookup(str(univ.OctetString('.'.join(['%s' % x for x in oid]))))
            except KeyError:
                offset = searchRecordByOid(oid, text, self.__textParser)
                subtreeFlag = exactMatch = False
            else:
                offset, subtreeFlag, prevOffset = line.split(str2octs(','), 2)
                subtreeFlag, exactMatch = int(subtreeFlag), True

            offset = int(offset)

            text.seek(offset)

            varsRemaining -= 1

            line, _, _ = getRecord(text)  # matched line
 
            while True:
                if exactMatch:
                    if context.get('nextFlag') and not subtreeFlag:
                        _nextLine, _, _ = getRecord(text)  # next line
                        if _nextLine:
                            _nextOid, _ = self.__textParser.evaluate(_nextLine, oidOnly=True)
                            try:
                                _, subtreeFlag, _ = self.__recordIndex.lookup(str(_nextOid)).split(str2octs(','), 2)
                            except KeyError:
                                log.error('data error for %s at %s, index broken?' % (self, _nextOid))
                                line = ''  # fatal error
                            else:
                                subtreeFlag = int(subtreeFlag)
                                line = _nextLine
                        else:
                            line = _nextLine
                else:  # search function above always rounds up to the next OID
                    if line:
                        _oid, _ = self.__textParser.evaluate(
                            line, oidOnly=True
                        )
                    else:  # eom
                        _oid = 'last'

                    try:
                        _, _, _prevOffset = self.__recordIndex.lookup(str(_oid)).split(str2octs(','), 2)
                    except KeyError:
                        log.error('data error for %s at %s, index broken?' % (self, _oid))
                        line = ''  # fatal error
                    else:
                        _prevOffset = int(_prevOffset)

                        # previous line serves a subtree?
                        if _prevOffset >= 0:
                            text.seek(_prevOffset)
                            _prevLine, _, _ = getRecord(text)
                            _prevOid, _ = self.__textParser.evaluate(
                                _prevLine, oidOnly=True
                            )
                            if _prevOid.isPrefixOf(oid):
                                # use previous line to the matched one
                                line = _prevLine
                                subtreeFlag = True

                if not line:
                    _oid = oid
                    _val = errorStatus
                    break

                callContext = context.copy()
                callContext.update(
                    (),
                    origOid=oid, 
                    origValue=val,
                    dataFile=self.__textFile,
                    subtreeFlag=subtreeFlag,
                    exactMatch=exactMatch,
                    errorStatus=errorStatus,
                    varsTotal=varsTotal,
                    varsRemaining=varsRemaining,
                    variationModules=variationModules
                )
 
                try:
                    _oid, _val = self.__textParser.evaluate(line, **callContext)
                    if _val is exval.endOfMib:
                        exactMatch = True
                        subtreeFlag = False
                        continue
                except NoDataNotification:
                    raise
                except MibOperationError:
                    raise
                except Exception:
                    _oid = oid
                    _val = errorStatus
                    log.error('data error at %s for %s: %s' % (self, textOid, sys.exc_info()[1]))

                break

            rspVarBinds.append((_oid, _val))

        log.info('Response var-binds: %s' % (', '.join(['%s=<%s>' % (vb[0], vb[1].prettyPrint()) for vb in rspVarBinds])))

        return rspVarBinds
 
    def __str__(self):
        return '%s controller' % self.__textFile


# Collect data files

def getDataFiles(tgtDir, topLen=None):
    if topLen is None:
        topLen = len(tgtDir.split(os.path.sep))
    dirContent = []
    for dFile in os.listdir(tgtDir):
        fullPath = tgtDir + os.path.sep + dFile
        inode = os.lstat(fullPath)
        if stat.S_ISLNK(inode.st_mode):
            relPath = fullPath.split(os.path.sep)[topLen:]
            fullPath = os.readlink(fullPath)
            if not os.path.isabs(fullPath):
                fullPath = tgtDir + os.path.sep + fullPath
            inode = os.stat(fullPath)
        else:
            relPath = fullPath.split(os.path.sep)[topLen:]
        if stat.S_ISDIR(inode.st_mode):
            dirContent = dirContent + getDataFiles(fullPath, topLen)
            continue
        if not stat.S_ISREG(inode.st_mode):
            continue
        dExt = os.path.splitext(dFile)[1][1:]
        if dExt not in recordSet:
            continue
        dirContent.append(
            (fullPath,
             recordSet[dExt],
             os.path.splitext(os.path.join(*relPath))[0].replace(os.path.sep, '/'))
        )
    return dirContent


# Lightweignt MIB instrumentation (API-compatible with pysnmp's)

class MibInstrumController:
    def __init__(self, dataFile):
        self.__dataFile = dataFile

    def __str__(self):
        return str(self.__dataFile)

    def __getCallContext(self, acInfo, nextFlag=False, setFlag=False):
        if acInfo is None:
            return {'nextFlag': nextFlag,
                    'setFlag': setFlag}

        acFun, snmpEngine = acInfo  # we injected snmpEngine object earlier
        # this API is first introduced in pysnmp 4.2.6 
        execCtx = snmpEngine.observer.getExecutionContext(
                'rfc3412.receiveMessage:request'
        )
        (transportDomain,
         transportAddress,
         securityModel,
         securityName,
         securityLevel,
         contextName,
         pduType) = (execCtx['transportDomain'],
                     execCtx['transportAddress'],
                     execCtx['securityModel'],
                     execCtx['securityName'],
                     execCtx['securityLevel'],
                     execCtx['contextName'],
                     execCtx['pdu'].getTagSet())

        log.info('SNMP EngineID %s, transportDomain %s, transportAddress %s, securityModel %s, '
                 'securityName %s, securityLevel %s' % (
            hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>',
            transportDomain, transportAddress, securityModel, securityName, securityLevel))

        return {'snmpEngine': snmpEngine,
                'transportDomain': transportDomain,
                'transportAddress': transportAddress,
                'securityModel': securityModel,
                'securityName': securityName,
                'securityLevel': securityLevel,
                'contextName': contextName,
                'nextFlag': nextFlag,
                'setFlag': setFlag}

    def readVars(self, varBinds, acInfo=None):
        return self.__dataFile.processVarBinds(
                varBinds, **self.__getCallContext(acInfo, False)
        )

    def readNextVars(self, varBinds, acInfo=None):
        return self.__dataFile.processVarBinds(
                varBinds, **self.__getCallContext(acInfo, True)
        )

    def writeVars(self, varBinds, acInfo=None):
        return self.__dataFile.processVarBinds(
                varBinds, **self.__getCallContext(acInfo, False, True)
        )


# Data files index as a MIB instrumentation in a dedicated SNMP context

class DataIndexInstrumController:
    indexSubOid = (1,)

    def __init__(self, baseOid=(1, 3, 6, 1, 4, 1, 20408, 999)):
        self.__db = indices.OidOrderedDict()
        self.__indexOid = baseOid + self.indexSubOid
        self.__idx = 1

    def __str__(self):
        return '<index> controller'

    def readVars(self, varBinds, acInfo=None):
        return [(vb[0], self.__db.get(vb[0], exval.noSuchInstance)) for vb in varBinds]

    def __getNextVal(self, key, default):
        try:
            key = self.__db.nextKey(key)
        except KeyError:
            return key, default
        else:
            return key, self.__db[key]
                                                            
    def readNextVars(self, varBinds, acInfo=None):
        return [self.__getNextVal(vb[0], exval.endOfMib) for vb in varBinds]

    def writeVars(self, varBinds, acInfo=None):
        return [(vb[0], exval.noSuchInstance) for vb in varBinds]
    
    def addDataFile(self, *args):
        for idx in range(len(args)):
            self.__db[
                self.__indexOid + (idx+1, self.__idx)
                ] = rfc1902.OctetString(args[idx])
        self.__idx += 1

mibInstrumControllerSet = {
    DataFile.layout: MibInstrumController
}


# Suggest variations of context name based on request data
def probeContext(transportDomain, transportAddress, contextName):
    candidate = [
        contextName, '.'.join([str(x) for x in transportDomain])
    ]
    if transportDomain[:len(udp.domainName)] == udp.domainName:
        candidate.append(transportAddress[0])
    elif udp6 and transportDomain[:len(udp6.domainName)] == udp6.domainName:
        candidate.append(
            str(transportAddress[0]).replace(':', '_')
        )
    elif unix and transportDomain[:len(unix.domainName)] == unix.domainName:
        candidate.append(transportAddress)

    candidate = [str(x) for x in candidate if x]

    while candidate:
        yield rfc1902.OctetString(os.path.normpath(os.path.sep.join(candidate)).replace(os.path.sep, '/')).asOctets()
        del candidate[-1]
 
# main script body starts here

helpMessage = """\
Usage: %s [--help]
    [--version ]
    [--debug=<%s>]
    [--debug-asn1=<%s>]
    [--daemonize]
    [--process-user=<uname>] [--process-group=<gname>]
    [--pid-file=<file>]
    [--logging-method=<%s[:args>]>]
    [--log-level=<%s>]
    [--cache-dir=<dir>]
    [--variation-modules-dir=<dir>]
    [--variation-module-options=<module[=alias][:args]>] 
    [--force-index-rebuild]
    [--validate-data]
    [--args-from-file=<file>]
    [--transport-id-offset=<number>]
    [--v2c-arch]
    [--v3-only]
    [--v3-engine-id=<hexvalue>]
    [--v3-context-engine-id=<hexvalue>]
    [--v3-user=<username>]
    [--v3-auth-key=<key>]
    [--v3-auth-proto=<%s>]
    [--v3-priv-key=<key>]
    [--v3-priv-proto=<%s>]
    [--data-dir=<dir>]
    [--max-varbinds=<number>]
    [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>]
    [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>]
    [--agent-unix-endpoint=</path/to/named/pipe>]""" % (
        sys.argv[0],
        '|'.join([x for x in getattr(pysnmp_debug, 'FLAG_MAP', getattr(pysnmp_debug, 'flagMap', ()))
                  if x != 'mibview']),
        '|'.join([x for x in getattr(pyasn1_debug, 'FLAG_MAP', getattr(pyasn1_debug, 'flagMap', ()))]),
        '|'.join(log.METHODS_MAP),
        '|'.join(log.LEVELS_MAP),
        '|'.join(sorted([x for x in authProtocols if x != 'NONE'])),
        '|'.join(sorted([x for x in privProtocols if x != 'NONE']))
    )

try:
    opts, params = getopt.getopt(sys.argv[1:], 'hv', [
        'help', 'version', 'debug=', 'debug-snmp=', 'debug-asn1=', 'daemonize',
        'process-user=', 'process-group=', 'pid-file=', 'logging-method=',
        'log-level=', 'device-dir=', 'cache-dir=', 'variation-modules-dir=',
        'force-index-rebuild', 'validate-device-data', 'validate-data',
        'v2c-arch', 'v3-only', 'transport-id-offset=', 
        'variation-module-options=',
        'args-from-file=',
        # this option starts new SNMPv3 engine configuration
        'v3-engine-id=',
        'v3-context-engine-id=',
        'v3-user=',
        'v3-auth-key=', 'v3-auth-proto=', 
        'v3-priv-key=', 'v3-priv-proto=',
        'data-dir=',
        'max-varbinds=',
        'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint='
    ])

except Exception:
    sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
    sys.exit(-1)

if params:
    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
    sys.exit(-1)

v3Args = []

for opt in opts:
    if opt[0] == '-h' or opt[0] == '--help':
        sys.stderr.write("""\
Synopsis:
  SNMP Agents Simulation tool. Responds to SNMP requests, variate responses
  based on transport addresses, SNMP community name or SNMPv3 context name.
  Can implement highly complex behavior through variation modules.
Documentation:
  http://snmplabs.com/snmpsim/simulating-agents.html
%s
""" % helpMessage)
        sys.exit(-1)
    if opt[0] == '-v' or opt[0] == '--version':
        import snmpsim
        import pysnmp
        import pyasn1

        sys.stderr.write("""\
SNMP Simulator version %s, written by Ilya Etingof <etingof@gmail.com>
Using foundation libraries: pysnmp %s, pyasn1 %s.
Python interpreter: %s
Software documentation and support at http://snmplabs.com/snmpsim
%s
""" % (snmpsim.__version__, hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown', hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown', sys.version, helpMessage))
        sys.exit(-1)
    elif opt[0] in ('--debug', '--debug-snmp'):
        pysnmp_debug.setLogger(pysnmp_debug.Debug(*opt[1].split(','), **dict(loggerName='snmpsimd.pysnmp')))
    elif opt[0] == '--debug-asn1':
        pyasn1_debug.setLogger(pyasn1_debug.Debug(*opt[1].split(','), **dict(loggerName='snmpsimd.pyasn1')))
    elif opt[0] == '--daemonize':
        foregroundFlag = False
    elif opt[0] == '--process-user':
        procUser = opt[1]
    elif opt[0] == '--process-group':
        procGroup = opt[1]
    elif opt[0] == '--pid-file':
        pidFile = opt[1]
    elif opt[0] == '--logging-method':
        loggingMethod = opt[1].split(':')
    elif opt[0] == '--log-level':
        loggingLevel = opt[1]
    elif opt[0] in ('--device-dir', '--data-dir'):
        if [x for x in v3Args if x[0] in ('--v3-engine-id',
                                          '--v3-context-engine-id')]:
            v3Args.append(opt)
        else:
            confdir.data.insert(0, opt[1])
    elif opt[0] == '--cache-dir':
        confdir.cache = opt[1]
    elif opt[0] == '--variation-modules-dir':
        confdir.variation.insert(0, opt[1])
    elif opt[0] == '--variation-module-options':
        args = opt[1].split(':', 1)
        try:
            modName, args = args[0], args[1]
        except:
            sys.stderr.write('ERROR: improper variation module options: %s\r\n%s\r\n' % (opt[1], helpMessage))
            sys.exit(-1)
        if '=' in modName:
            modName, alias = modName.split('=', 1)
        else:
            alias = os.path.splitext(os.path.basename(modName))[0]
        if modName not in variationModulesOptions:
            variationModulesOptions[modName] = []
        variationModulesOptions[modName].append((alias, args))
    elif opt[0] == '--force-index-rebuild':
        forceIndexBuild = True
    elif opt[0] in ('--validate-device-data', '--validate-data'):
        validateData = True
    elif opt[0] == '--v2c-arch':
        v2cArch = True
    elif opt[0] == '--v3-only':
        v3Only = True
    elif opt[0] == '--transport-id-offset':
        try:
            transportIdOffset = max(0, int(opt[1]))
        except:
            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    # processing of the following args is postponed till SNMP engine startup
    elif opt[0] in ('--agent-udpv4-endpoint',
                    '--agent-udpv6-endpoint',
                    '--agent-unix-endpoint'):
        v3Args.append(opt)
    elif opt[0] in ('--agent-udpv4-endpoints-list',
                    '--agent-udpv6-endpoints-list',
                    '--agent-unix-endpoints-list'):
        sys.stderr.write('ERROR: use --args-from-file=<file> option to list many endpoints\r\n%s\r\n' % helpMessage)
        sys.exit(-1)
    elif opt[0] in ('--v3-engine-id', '--v3-context-engine-id',
                    '--v3-user', '--v3-auth-key', '--v3-auth-proto',
                    '--v3-priv-key', '--v3-priv-proto'):
        v3Args.append(opt)
    elif opt[0] == '--max-varbinds':
        try:
            if '--v3-engine-id' in [x[0] for x in v3Args]:
                v3Args.append((opt[0], max(1, int(opt[1]))))
            else:
                maxVarBinds = max(1, int(opt[1]))
        except:
            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    elif opt[0] == '--args-from-file':
        try:
            v3Args.extend(
                [x.split('=', 1) for x in open(opt[1]).read().split()]
            )
        except:
            sys.stderr.write('ERROR: file %s opening failure: %s\r\n%s\r\n' % (opt[1], sys.exc_info()[1], helpMessage))
            sys.exit(-1)

if v2cArch and (v3Only or [x for x in v3Args if x[0][:4] == '--v3']):
    sys.stderr.write('ERROR: either of --v2c-arch or --v3-* options should be used\r\n%s\r\n' % helpMessage)
    sys.exit(-1)

for opt in tuple(v3Args):
    if opt[0] == '--agent-udpv4-endpoint':
        try:
            v3Args.append((opt[0], IPv4TransportEndpoints().add(opt[1])))
        except:
            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    elif opt[0] == '--agent-udpv6-endpoint':
        try:
            v3Args.append((opt[0], IPv6TransportEndpoints().add(opt[1])))
        except:
            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    elif opt[0] == '--agent-unix-endpoint':
        try:
            v3Args.append((opt[0], UnixTransportEndpoints().add(opt[1])))
        except:
            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
            sys.exit(-1)
    else:
        v3Args.append(opt) 

if v3Args[:len(v3Args)//2] == v3Args[len(v3Args)//2:]:
    sys.stderr.write('ERROR: agent endpoint address(es) not specified\r\n%s\r\n' % helpMessage)
    sys.exit(-1)
else:
    v3Args = v3Args[len(v3Args)//2:]

with daemon.PrivilegesOf(procUser, procGroup):

    try:
        log.setLogger(PROGRAM_NAME, *loggingMethod, force=True)

        if loggingLevel:
            log.setLevel(loggingLevel)

    except SnmpsimError:
        sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
        sys.exit(1)

if not foregroundFlag:
    try:
        daemon.daemonize(pidFile)
    except:
        sys.stderr.write('ERROR: cant daemonize process: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
        sys.exit(-1)

# hook up variation modules

for variationModulesDir in confdir.variation:
    log.info('Scanning "%s" directory for variation modules...' % variationModulesDir)
    if not os.path.exists(variationModulesDir):
        log.info('Directory "%s" does not exist' % variationModulesDir)
        continue
    for dFile in os.listdir(variationModulesDir):
        if dFile[-3:] != '.py':
            continue
        _toLoad = []
        modName = os.path.splitext(os.path.basename(dFile))[0]
        if modName in variationModulesOptions:
            while variationModulesOptions[modName]:
                alias, args = variationModulesOptions[modName].pop()
                _toLoad.append((alias, args))
            del variationModulesOptions[modName]
        else:
            _toLoad.append((modName, ''))

        mod = os.path.abspath(variationModulesDir + os.path.sep + dFile)

        for alias, args in _toLoad:
            if alias in variationModules:
                log.error('ignoring duplicate variation module "%s" at "%s"' % (alias, mod))
                continue

            ctx = {
                'path': mod,
                'alias': alias,
                'args': args,
                'moduleContext': {}
            }

            try:
                if sys.version_info[0] > 2:
                    exec(compile(open(mod).read(), mod, 'exec'), ctx)
                else:
                    execfile(mod, ctx)
            except Exception:
                log.error('Variation module "%s" execution failure: %s' % (mod, sys.exc_info()[1]))
                sys.exit(-1)
            else:
                # moduleContext, agentContexts, recordContexts
                variationModules[alias] = ctx, {}, {}

    log.info('A total of %s modules found in %s' % (len(variationModules), variationModulesDir))

if variationModulesOptions:
    log.msg('WARNING: unused options for variation modules: %s' % ', '.join(variationModulesOptions.keys()))

if not os.path.exists(confdir.cache):
    try:
        os.makedirs(confdir.cache)
    except OSError:
        log.error('failed to create cache directory "%s": %s' % (confdir.cache, sys.exc_info()[1]))
        sys.exit(-1)
    else:
        log.info('Cache directory "%s" created' % confdir.cache)

if variationModules:
    log.info('Initializing variation modules...')
    for name, modulesContexts in variationModules.items():
        body = modulesContexts[0]
        for x in ('init', 'variate', 'shutdown'):
            if x not in body:
                log.error('missing "%s" handler at variation module "%s"' % (x, name))
                sys.exit(-1)
        try:
            body['init'](options=body['args'], mode='variating')

        except Exception:
            log.error('Variation module "%s" from "%s" load FAILED: %s' % (
                body['alias'], body['path'], sys.exc_info()[1]))

        else:
            log.info('Variation module "%s" from "%s" loaded OK' % (body['alias'], body['path']))


# Build pysnmp Managed Objects base from data files information

def configureManagedObjects(dataDirs, dataIndexInstrumController,
                            snmpEngine=None, snmpContext=None):
    _mibInstrums = {}
    _dataFiles = {}

    for dataDir in dataDirs:
        log.info('Scanning "%s" directory for %s data files...' % (dataDir, ','.join([' *%s%s' % (os.path.extsep, x.ext) for x in recordSet.values()])))
        if not os.path.exists(dataDir):
            log.info('Directory "%s" does not exist' % dataDir)
            continue
        log.msg.incIdent()
        for fullPath, textParser, communityName in getDataFiles(dataDir):
            if communityName in _dataFiles:
                log.error('ignoring duplicate Community/ContextName "%s" for data file %s '
                          '(%s already loaded)' % (communityName, fullPath, _dataFiles[communityName]))
                continue
            elif fullPath in _mibInstrums:
                mibInstrum = _mibInstrums[fullPath]
                log.info('Configuring *shared* %s' % (mibInstrum,))
            else:
                dataFile = DataFile(fullPath, textParser).indexText(forceIndexBuild)
                mibInstrum = mibInstrumControllerSet[dataFile.layout](dataFile)

                _mibInstrums[fullPath] = mibInstrum
                _dataFiles[communityName] = fullPath

                log.info('Configuring %s' % (mibInstrum,))

            log.info('SNMPv1/2c community name: %s' % (communityName,))

            if v2cArch:
                contexts[univ.OctetString(communityName)] = mibInstrum
            
                dataIndexInstrumController.addDataFile(
                    fullPath, communityName
                )
            else:
                agentName = contextName = md5(univ.OctetString(communityName).asOctets()).hexdigest()

                if not v3Only:
                    # snmpCommunityTable::snmpCommunityIndex can't be > 32
                    config.addV1System(
                        snmpEngine, agentName, communityName, contextName=contextName
                    )

                snmpContext.registerContextName(contextName, mibInstrum)
                
                if len(communityName) <= 32:
                    snmpContext.registerContextName(communityName, mibInstrum)
                         
                dataIndexInstrumController.addDataFile(
                    fullPath, communityName, contextName
                )
                         
                log.info('SNMPv3 Context Name: %s%s' % (contextName, len(communityName) <= 32 and ' or %s' % communityName or ''))

        log.msg.decIdent()

    del _mibInstrums
    del _dataFiles


if v2cArch:
    def getBulkHandler(reqVarBinds, nonRepeaters, maxRepetitions, readNextVars):
        N = min(int(nonRepeaters), len(reqVarBinds))
        M = int(maxRepetitions)
        R = max(len(reqVarBinds)-N, 0)
        if R:
            M = min(M, maxVarBinds/R)

        if N:
            rspVarBinds = readNextVars(reqVarBinds[:N])
        else:
            rspVarBinds = []

        varBinds = reqVarBinds[-R:]
        while M and R:
            rspVarBinds.extend(
                readNextVars(varBinds)
            )
            varBinds = rspVarBinds[-R:]
            M -= 1

        return rspVarBinds

    def commandResponderCbFun(transportDispatcher, transportDomain,
                              transportAddress, wholeMsg):
        while wholeMsg:
            msgVer = api.decodeMessageVersion(wholeMsg)
            if msgVer in api.protoModules:
                pMod = api.protoModules[msgVer]
            else:
                log.error('Unsupported SNMP version %s' % (msgVer,))
                return
            reqMsg, wholeMsg = decoder.decode(
                wholeMsg, asn1Spec=pMod.Message(),
                )

            communityName = reqMsg.getComponentByPosition(1)
            for candidate in probeContext(transportDomain, transportAddress, communityName):
                if candidate in contexts:
                    log.info('Using %s selected by candidate %s; transport ID %s, source address %s, '
                             'community name "%s"' % (
                        contexts[candidate], candidate, univ.ObjectIdentifier(transportDomain),
                        transportAddress[0], communityName))
                    communityName = candidate
                    break
            else:
                log.error('No data file selected for transport ID %s, source address %s, '
                          'community name "%s"' % (
                    univ.ObjectIdentifier(transportDomain), transportAddress[0], communityName))
                return wholeMsg
            
            rspMsg = pMod.apiMessage.getResponse(reqMsg)
            rspPDU = pMod.apiMessage.getPDU(rspMsg)        
            reqPDU = pMod.apiMessage.getPDU(reqMsg)
    
            if reqPDU.isSameTypeWith(pMod.GetRequestPDU()):
                backendFun = contexts[communityName].readVars
            elif reqPDU.isSameTypeWith(pMod.SetRequestPDU()):
                backendFun = contexts[communityName].writeVars
            elif reqPDU.isSameTypeWith(pMod.GetNextRequestPDU()):
                backendFun = contexts[communityName].readNextVars
            elif (hasattr(pMod, 'GetBulkRequestPDU') and
                    reqPDU.isSameTypeWith(pMod.GetBulkRequestPDU())):
                if not msgVer:
                    log.info('GETBULK over SNMPv1 from %s:%s' % (transportDomain, transportAddress))
                    return wholeMsg

                def backendFun(varBinds):
                    return getBulkHandler(
                        varBinds, pMod.apiBulkPDU.getNonRepeaters(reqPDU),
                        pMod.apiBulkPDU.getMaxRepetitions(reqPDU),
                        contexts[communityName].readNextVars
                    )
            else:
                log.error('Unsupported PDU type %s from %s:%s' % (reqPDU.__class__.__name__, transportDomain, transportAddress))
                return wholeMsg
   
            try: 
                varBinds = backendFun(
                    pMod.apiPDU.getVarBinds(reqPDU)
                )

            except NoDataNotification:
                return wholeMsg

            except Exception:
                log.error('Ignoring SNMP engine failure: %s' % sys.exc_info()[1])

            # Poor man's v2c->v1 translation
            errorMap = {
                rfc1902.Counter64.tagSet: 5,
                rfc1905.NoSuchObject.tagSet: 2,
                rfc1905.NoSuchInstance.tagSet: 2,
                rfc1905.EndOfMibView.tagSet: 2
            }
 
            if not msgVer:
                for idx in range(len(varBinds)):
                    oid, val = varBinds[idx]
                    if val.tagSet in errorMap:
                        varBinds = pMod.apiPDU.getVarBinds(reqPDU)
                        pMod.apiPDU.setErrorStatus(rspPDU, errorMap[val.tagSet])
                        pMod.apiPDU.setErrorIndex(rspPDU, idx + 1)
                        break

            pMod.apiPDU.setVarBinds(rspPDU, varBinds)
            
            transportDispatcher.sendMessage(
                encoder.encode(rspMsg), transportDomain, transportAddress
            )
            
        return wholeMsg

else:  # v3arch
    def probeHashContext(self, snmpEngine, stateReference, contextName):
        # this API is first introduced in pysnmp 4.2.6 
        execCtx = snmpEngine.observer.getExecutionContext('rfc3412.receiveMessage:request')

        transportDomain, transportAddress = (
            execCtx['transportDomain'], execCtx['transportAddress']
        )

        for candidate in probeContext(transportDomain, transportAddress, contextName):
            if len(candidate) > 32:
                probedContextName = md5(candidate).hexdigest()
            else:
                probedContextName = candidate
            try:
                mibInstrum = self.snmpContext.getMibInstrum(probedContextName)
            except error.PySnmpError:
                pass
            else:
                log.info('Using %s selected by candidate %s; transport ID %s, source address %s, context name "%s"' % (mibInstrum, candidate, univ.ObjectIdentifier(transportDomain), transportAddress[0], probedContextName))
                contextName = probedContextName
                break
        else:
            mibInstrum = self.snmpContext.getMibInstrum(contextName)
            log.info('Using %s selected by contextName "%s", transport ID %s, source address %s' % (mibInstrum, contextName, univ.ObjectIdentifier(transportDomain), transportAddress[0]))

        if not isinstance(mibInstrum, (MibInstrumController, DataIndexInstrumController)):
            log.error('LCD access denied (contextName does not match any data file)')
            raise NoDataNotification()

        return contextName

    class GetCommandResponder(cmdrsp.GetCommandResponder):
        def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
            try:
                cmdrsp.GetCommandResponder.handleMgmtOperation(
                    self, snmpEngine, stateReference, 
                    probeHashContext(
                        self, snmpEngine, stateReference, contextName
                    ),
                    PDU, (None, snmpEngine)  # custom acInfo
                )
            except NoDataNotification:
                self.releaseStateInformation(stateReference)

    class SetCommandResponder(cmdrsp.SetCommandResponder):
        def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
            try:
                cmdrsp.SetCommandResponder.handleMgmtOperation(
                    self, snmpEngine, stateReference, 
                    probeHashContext(
                        self, snmpEngine, stateReference, contextName
                    ),
                    PDU, (None, snmpEngine)  # custom acInfo
                )
            except NoDataNotification:
                self.releaseStateInformation(stateReference)

    class NextCommandResponder(cmdrsp.NextCommandResponder):
        def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
            try:
                cmdrsp.NextCommandResponder.handleMgmtOperation(
                    self, snmpEngine, stateReference, 
                    probeHashContext(
                        self, snmpEngine, stateReference, contextName
                    ),
                    PDU, (None, snmpEngine)  # custom acInfo
                )
            except NoDataNotification:
                self.releaseStateInformation(stateReference)

    class BulkCommandResponder(cmdrsp.BulkCommandResponder):
        def handleMgmtOperation(self, snmpEngine, stateReference, contextName, PDU, acInfo):
            try:
                cmdrsp.BulkCommandResponder.handleMgmtOperation(
                    self, snmpEngine, stateReference, 
                    probeHashContext(
                        self, snmpEngine, stateReference, contextName
                    ),
                    PDU, (None, snmpEngine)  # custom acInfo
                )
            except NoDataNotification:
                self.releaseStateInformation(stateReference)

# Start configuring SNMP engine(s)

transportDispatcher = AsynsockDispatcher()

if v2cArch:
    # Configure access to data index
   
    dataIndexInstrumController = DataIndexInstrumController()
 
    contexts = {univ.OctetString('index'): dataIndexInstrumController}

    configureManagedObjects(confdir.data, dataIndexInstrumController)

    contexts['index'] = dataIndexInstrumController
    
    agentUDPv4Endpoints = []
    agentUDPv6Endpoints = []
    agentUnixEndpoints = []

    for opt in v3Args:
        if opt[0] == '--agent-udpv4-endpoint':
            agentUDPv4Endpoints.append(opt[1])
        elif opt[0] == '--agent-udpv6-endpoint':
            agentUDPv6Endpoints.append(opt[1])
        elif opt[0] == '--agent-unix-endpoint':
            agentUnixEndpoints.append(opt[1])

    if (not agentUDPv4Endpoints and
            not agentUDPv6Endpoints and
            not agentUnixEndpoints):
        log.error('agent endpoint address(es) not specified')
        sys.exit(-1)

    log.info('Maximum number of variable bindings in SNMP response: %s' % maxVarBinds)

    # Configure socket server
   
    transportIndex = transportIdOffset
    for agentUDPv4Endpoint in agentUDPv4Endpoints:
        transportDomain = udp.domainName + (transportIndex,)
        transportIndex += 1
        transportDispatcher.registerTransport(
                transportDomain, agentUDPv4Endpoint[0]
        )
        log.info('Listening at UDP/IPv4 endpoint %s, transport ID %s' % (agentUDPv4Endpoint[1], '.'.join([str(x) for x in transportDomain])))

    transportIndex = transportIdOffset
    for agentUDPv6Endpoint in agentUDPv6Endpoints:
        transportDomain = udp6.domainName + (transportIndex,)
        transportIndex += 1
        transportDispatcher.registerTransport(
                transportDomain, agentUDPv4Endpoint[0]
        )
        log.info('Listening at UDP/IPv6 endpoint %s, transport ID %s' % (agentUDPv4Endpoint[1], '.'.join([str(x) for x in transportDomain])))

    transportIndex = transportIdOffset
    for agentUnixEndpoint in agentUnixEndpoints:
        transportDomain = unix.domainName + (transportIndex,)
        transportIndex += 1
        transportDispatcher.registerTransport(
                transportDomain, agentUnixEndpoint[0]
        )
        log.info('Listening at UNIX domain socket endpoint %s, transport ID %s' % (agentUnixEndpoint[1], '.'.join([str(x) for x in transportDomain])))

    transportDispatcher.registerRecvCbFun(commandResponderCbFun)
 
else:  # v3 mode
    if hasattr(transportDispatcher, 'registerRoutingCbFun'):
        transportDispatcher.registerRoutingCbFun(lambda td, t, d: td)
    else:
        log.info('WARNING: upgrade pysnmp to 4.2.5 or later get multi-engine ID feature working!')

    if v3Args and v3Args[0][0] != '--v3-engine-id':
        v3Args.insert(0, ('--v3-engine-id', 'auto'))
    v3Args.append(('end-of-options', ''))

    def registerTransportDispatcher(snmpEngine, transportDispatcher,
                                    transportDomain):
        if hasattr(transportDispatcher, 'registerRoutingCbFun'):
            snmpEngine.registerTransportDispatcher(
                transportDispatcher, transportDomain
            )
        else:
            try:
                snmpEngine.registerTransportDispatcher(
                    transportDispatcher
                )
            except error.PySnmpError:
                log.msg('WARNING: upgrade pysnmp to 4.2.5 or later get multi-engine ID feature working!')
                raise

    snmpEngine = None
    transportIndex = {
        'udpv4': transportIdOffset,
        'udpv6': transportIdOffset,
        'unix': transportIdOffset
    }

    for opt in v3Args:
        if opt[0] in ('--v3-engine-id', 'end-of-options'):
            if snmpEngine:
                log.info('--- SNMP Engine configuration')

                log.info('SNMPv3 EngineID: %s' % (hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>',))

                if not v3ContextEngineIds:
                    v3ContextEngineIds.append((None, []))

                log.msg.incIdent()

                log.info('--- Data directories configuration')

                for v3ContextEngineId, ctxDataDirs in v3ContextEngineIds:
                    snmpContext = context.SnmpContext(
                        snmpEngine, v3ContextEngineId
                    )

                    log.msg('SNMPv3 Context Engine ID: %s' % snmpContext.contextEngineId.prettyPrint())

                    dataIndexInstrumController = DataIndexInstrumController()

                    configureManagedObjects(
                        ctxDataDirs or dataDirs or confdir.data, 
                        dataIndexInstrumController,
                        snmpEngine,
                        snmpContext
                    )

                # Configure access to data index

                config.addV1System(snmpEngine, 'index',
                                   'index', contextName='index')

                log.info('--- SNMPv3 USM configuration')

                if not v3Users:
                    v3Users = ['simulator']
                    v3AuthKeys[v3Users[0]] = 'auctoritas'
                    v3AuthProtos[v3Users[0]] = 'MD5'
                    v3PrivKeys[v3Users[0]] = 'privatus'
                    v3PrivProtos[v3Users[0]] = 'DES'

                for v3User in v3Users:
                    if v3User in v3AuthKeys:
                        if v3User not in v3AuthProtos:
                            v3AuthProtos[v3User] = 'MD5'
                    elif v3User in v3AuthProtos:
                        log.error('auth protocol configured without key for user %s' % v3User)
                        sys.exit(-1)
                    else:
                        v3AuthKeys[v3User] = None
                        v3AuthProtos[v3User] = 'NONE'
                    if v3User in v3PrivKeys:
                        if v3User not in v3PrivProtos:
                            v3PrivProtos[v3User] = 'DES'
                    elif v3User in v3PrivProtos:
                        log.error('privacy protocol configured without key for user %s' % v3User)
                        sys.exit(-1)
                    else:
                        v3PrivKeys[v3User] = None
                        v3PrivProtos[v3User] = 'NONE'
                    if authProtocols[v3AuthProtos[v3User]] == config.usmNoAuthProtocol and privProtocols[v3PrivProtos[v3User]] != config.usmNoPrivProtocol:
                        log.error('privacy impossible without authentication for USM user %s' % v3User)
                        sys.exit(-1)

                    try:
                        config.addV3User(
                            snmpEngine,
                            v3User,
                            authProtocols[v3AuthProtos[v3User]],
                            v3AuthKeys[v3User],
                            privProtocols[v3PrivProtos[v3User]], 
                            v3PrivKeys[v3User]
                        )
                    except error.PySnmpError:
                        log.error('bad USM values for user %s: %s' % (v3User, sys.exc_info()[1]))
                        sys.exit(-1)

                    log.info('SNMPv3 USM SecurityName: %s' % v3User)

                    if authProtocols[v3AuthProtos[v3User]] != config.usmNoAuthProtocol:
                        log.info('SNMPv3 USM authentication key: %s, authentication protocol: %s' % (
                            v3AuthKeys[v3User], v3AuthProtos[v3User]))
                    if privProtocols[v3PrivProtos[v3User]] != config.usmNoPrivProtocol:
                        log.info('SNMPv3 USM encryption (privacy) key: %s, encryption protocol: %s' % (
                            v3PrivKeys[v3User], v3PrivProtos[v3User]))

                snmpContext.registerContextName('index', dataIndexInstrumController)

                log.info('Maximum number of variable bindings in SNMP response: %s' % localMaxVarBinds)

                log.info('--- Transport configuration')

                if (not agentUDPv4Endpoints and
                        not agentUDPv6Endpoints and
                        not agentUnixEndpoints):
                    log.error('agent endpoint address(es) not specified for SNMP engine ID %s' % v3EngineId)
                    sys.exit(-1)

                for agentUDPv4Endpoint in agentUDPv4Endpoints:
                    transportDomain = udp.domainName + (transportIndex['udpv4'],)
                    transportIndex['udpv4'] += 1
                    registerTransportDispatcher(
                        snmpEngine, transportDispatcher, transportDomain
                    )
                    config.addSocketTransport(
                        snmpEngine,
                        transportDomain, agentUDPv4Endpoint[0]
                    )
                    log.info('Listening at UDP/IPv4 endpoint %s, transport ID %s' % (
                        agentUDPv4Endpoint[1], '.'.join([str(x) for x in transportDomain])))

                for agentUDPv6Endpoint in agentUDPv6Endpoints:
                    transportDomain = udp6.domainName + (transportIndex['udpv6'],)
                    transportIndex['udpv6'] += 1
                    registerTransportDispatcher(
                        snmpEngine, transportDispatcher, transportDomain
                    )
                    config.addSocketTransport(
                        snmpEngine,
                        transportDomain, agentUDPv6Endpoint[0]
                    )
                    log.info('Listening at UDP/IPv6 endpoint %s, transport ID %s' % (
                        agentUDPv6Endpoint[1], '.'.join([str(x) for x in transportDomain])))

                for agentUnixEndpoint in agentUnixEndpoints:
                    transportDomain = unix.domainName + (transportIndex['unix'],)
                    transportIndex['unix'] += 1
                    registerTransportDispatcher(
                        snmpEngine, transportDispatcher, transportDomain
                    )
                    config.addSocketTransport(
                        snmpEngine,
                        transportDomain, agentUnixEndpoint[0]
                    )
                    log.info('Listening at UNIX domain socket endpoint %s, transport ID %s' % (
                        agentUnixEndpoint[1], '.'.join([str(x) for x in transportDomain])))

                # SNMP applications
                GetCommandResponder(snmpEngine, snmpContext)
                SetCommandResponder(snmpEngine, snmpContext)
                NextCommandResponder(snmpEngine, snmpContext)
                BulkCommandResponder(snmpEngine, snmpContext).maxVarBinds = localMaxVarBinds

                log.msg.decIdent()

            if opt[0] == 'end-of-options':
                break

            # Prepare for next engine ID configuration

            v3ContextEngineIds = []
            dataDirs = []
            localMaxVarBinds = maxVarBinds
            v3Users = []
            v3AuthKeys = {}
            v3AuthProtos = {}
            v3PrivKeys = {}
            v3PrivProtos = {}
            agentUDPv4Endpoints = []
            agentUDPv6Endpoints = []
            agentUnixEndpoints = []

            try:
                v3EngineId = opt[1]
                if v3EngineId.lower() == 'auto':
                    snmpEngine = engine.SnmpEngine()
                else:
                    snmpEngine = engine.SnmpEngine(snmpEngineID=univ.OctetString(hexValue=v3EngineId))
            except:
                log.error('SNMPv3 Engine initialization failed, EngineID "%s": %s' % (v3EngineId, sys.exc_info()[1]))
                sys.exit(-1)

            config.addContext(snmpEngine, '')

        elif opt[0] == '--v3-context-engine-id':
            v3ContextEngineIds.append((univ.OctetString(hexValue=opt[1]), []))
        elif opt[0] == '--data-dir':
            if v3ContextEngineIds:
                v3ContextEngineIds[-1][1].append(opt[1])
            else:
                dataDirs.append(opt[1])
        elif opt[0] == '--max-varbinds':
            localMaxVarBinds = opt[1]
        elif opt[0] == '--v3-user':
            v3Users.append(opt[1])
        elif opt[0] == '--v3-auth-key':
            if not v3Users:
                log.error('--v3-user should precede %s' % opt[0])
                sys.exit(-1)
            if v3Users[-1] in v3AuthKeys:
                log.error('repetitive %s option for user %s' % (opt[0], v3Users[-1]))
                sys.exit(-1)
            v3AuthKeys[v3Users[-1]] = opt[1]
        elif opt[0] == '--v3-auth-proto':
            if opt[1].upper() not in authProtocols:
                log.error('bad v3 auth protocol %s' % opt[1])
                sys.exit(-1)
            else:
                if not v3Users:
                    log.error('--v3-user should precede %s' % opt[0])
                    sys.exit(-1)
                if v3Users[-1] in v3AuthProtos:
                    log.error('repetitive %s option for user %s' % (opt[0], v3Users[-1]))
                    sys.exit(-1)
                v3AuthProtos[v3Users[-1]] = opt[1].upper()
        elif opt[0] == '--v3-priv-key':
            if not v3Users:
                log.error('--v3-user should precede %s' % opt[0])
                sys.exit(-1)
            if v3Users[-1] in v3PrivKeys:
                log.error('repetitive %s option for user %s' % (opt[0], v3Users[-1]))
                sys.exit(-1)
            v3PrivKeys[v3Users[-1]] = opt[1]
        elif opt[0] == '--v3-priv-proto':
            if opt[1].upper() not in privProtocols:
                log.error('bad v3 privacy protocol %s' % opt[1])
                sys.exit(-1)
            else:
                if not v3Users:
                    log.error('--v3-user should precede %s' % opt[0])
                    sys.exit(-1)
                if v3Users[-1] in v3PrivProtos:
                    log.error('repetitive %s option for user %s' % (opt[0], v3Users[-1]))
                    sys.exit(-1)
                v3PrivProtos[v3Users[-1]] = opt[1].upper()
        elif opt[0] == '--agent-udpv4-endpoint':
            agentUDPv4Endpoints.append(opt[1])
        elif opt[0] == '--agent-udpv6-endpoint':
            agentUDPv6Endpoints.append(opt[1])
        elif opt[0] == '--agent-unix-endpoint':
            agentUnixEndpoints.append(opt[1])

# Run mainloop

transportDispatcher.jobStarted(1)  # server job would never finish

# Python 2.4 does not support the "finally" clause

exc_info = None

with daemon.PrivilegesOf(procUser, procGroup, final=True):

    try:
        transportDispatcher.runDispatcher()

    except KeyboardInterrupt:
        log.info('Shutting down process...')

    except Exception:
        exc_info = sys.exc_info()

    if variationModules:
        log.info('Shutting down variation modules:')

        for name, contexts in variationModules.items():
            body = contexts[0]
            try:
                body['shutdown'](options=body['args'], mode='variation')

            except Exception:
                log.error('Variation module "%s" shutdown FAILED: %s' % (name, sys.exc_info()[1]))

            else:
                log.info('Variation module "%s" shutdown OK' % name)

    transportDispatcher.closeDispatcher()

    log.info('Process terminated')

    if exc_info:
        for line in traceback.format_exception(*exc_info):
            log.error(line.replace('\n', ';'))
